# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import List, Union, AsyncGenerator, Callable

from openai.types.chat import ChatCompletionMessageParam
from mito_ai.completions.models import ChatMessageMetadata, MessageType, CompletionRequest, CompletionStreamChunk, CompletionReply
from mito_ai.completions.prompt_builders.chat_prompt import create_chat_prompt
from mito_ai.completions.providers import OpenAIProvider
from mito_ai.completions.message_history import GlobalMessageHistory
from mito_ai.completions.completion_handlers.completion_handler import CompletionHandler
from mito_ai.completions.completion_handlers.utils import append_chat_system_message, create_ai_optimized_message

__all__ = ["get_chat_completion", "stream_chat_completion"]

class ChatCompletionHandler(CompletionHandler[ChatMessageMetadata]):
    """Handler for chat completions."""
    
    @staticmethod
    async def get_completion(
        metadata: ChatMessageMetadata,
        provider: OpenAIProvider,
        message_history: GlobalMessageHistory,
        model: str
    ) -> str:
        """Get a chat completion from the AI provider."""

        if metadata.index is not None:
            message_history.truncate_histories(
                index=metadata.index,
                thread_id=metadata.threadId
            )

        # Add the system message if it doesn't alredy exist
        await append_chat_system_message(message_history, model, provider, metadata.threadId)
        
        # Create the prompt
        prompt = create_chat_prompt(
            metadata.variables or [], 
            metadata.files or [],
            metadata.activeCellCode, 
            metadata.activeCellId,
            metadata.base64EncodedActiveCellOutput is not None and metadata.base64EncodedActiveCellOutput != '',
            metadata.input,
            metadata.selectedRules
        )
        display_prompt = f"```python{metadata.activeCellCode or ''}```{metadata.input}"
        
        # Add the prompt to the message history
        new_ai_optimized_message = create_ai_optimized_message(prompt, metadata.base64EncodedActiveCellOutput)
        new_display_optimized_message: ChatCompletionMessageParam = {"role": "user", "content": display_prompt}
        await message_history.append_message(new_ai_optimized_message, new_display_optimized_message, model, provider, metadata.threadId)
        
        # Get the completion (non-streaming)
        completion = await provider.request_completions(
            messages=message_history.get_ai_optimized_history(metadata.threadId), 
            model=model,
            message_type=MessageType.CHAT,
            user_input=metadata.input,
            thread_id=metadata.threadId
        )
        
        ai_response_message: ChatCompletionMessageParam = {"role": "assistant", "content": completion}
        await message_history.append_message(ai_response_message, ai_response_message, model, provider, metadata.threadId)

        return completion
    
    @staticmethod
    async def stream_completion(
        metadata: ChatMessageMetadata,
        provider: OpenAIProvider,
        message_history: GlobalMessageHistory,
        message_id: str,
        reply_fn: Callable[[Union[CompletionReply, CompletionStreamChunk]], None],
        model: str
    ) -> str:
        """Stream chat completions from the AI provider.
        
        Args:
            metadata: The metadata for the chat completion request.
            provider: The AI provider to use.
            message_history: The message history for this conversation.
            message_id: The ID of the message being processed.
            reply_fn: Function to call with each chunk for streaming replies.
            
        Returns:
            The accumulated response string.
        """
        index = metadata.index

        if index is not None:
            message_history.truncate_histories(
                index=index,
                thread_id=metadata.threadId
            )
        
        # Add the system message if it doesn't already exist
        await append_chat_system_message(message_history, model, provider, metadata.threadId)
        
        # Create the prompt
        prompt = create_chat_prompt(
            metadata.variables or [], 
            metadata.files or [],
            metadata.activeCellCode, 
            metadata.activeCellId,
            metadata.base64EncodedActiveCellOutput is not None and metadata.base64EncodedActiveCellOutput != '',
            metadata.input,
            metadata.selectedRules
        )
        display_prompt = f"```python{metadata.activeCellCode or ''}```{metadata.input}"
        
        # Add the prompt to the message history
        new_ai_optimized_message = create_ai_optimized_message(prompt, metadata.base64EncodedActiveCellOutput)
        new_display_optimized_message: ChatCompletionMessageParam = {"role": "user", "content": display_prompt}
        await message_history.append_message(new_ai_optimized_message, new_display_optimized_message, model, provider, metadata.threadId)
        
        # Stream the completions using the provider's stream method
        accumulated_response = await provider.stream_completions(
            message_type=MessageType.CHAT,
            messages=message_history.get_ai_optimized_history(metadata.threadId),
            model=model,
            message_id=message_id,
            reply_fn=reply_fn,
            user_input=metadata.input,
            thread_id=metadata.threadId
        )

        # Save the accumulated response to message history
        ai_response_message: ChatCompletionMessageParam = {
            "role": "assistant",
            "content": accumulated_response,
        }
        await message_history.append_message(
            ai_response_message, ai_response_message, model, provider, metadata.threadId
        )

        return accumulated_response

# Use the static methods directly
get_chat_completion = ChatCompletionHandler.get_completion
stream_chat_completion = ChatCompletionHandler.stream_completion



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import json
import tornado
from jupyter_server.base.handlers import APIHandler
from mito_ai.settings.utils import (
    get_settings_field,
    set_settings_field,
    ensure_settings_file_exists,
)


class SettingsHandler(APIHandler):
    """Handler for operations on a specific setting"""

    @tornado.web.authenticated
    @ensure_settings_file_exists
    def get(self, key: str) -> None:
        """Get a specific setting by key"""
        setting_value = get_settings_field(key)
        if setting_value is None:
            self.set_status(404)
            self.finish(json.dumps({"error": f"Setting with key '{key}' not found"}))
        else:
            self.finish(json.dumps({"key": key, "value": setting_value}))

    @tornado.web.authenticated
    @ensure_settings_file_exists
    def put(self, key: str) -> None:
        """Update or create a specific setting"""
        data = json.loads(self.request.body)
        if "value" not in data:
            self.set_status(400)
            self.finish(json.dumps({"error": "Value is required"}))
            return

        set_settings_field(key, data["value"])
        self.finish(
            json.dumps({"status": "updated", "key": key, "value": data["value"]})
        )



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import List
from mito_ai.completions.prompt_builders.prompt_constants import (
    FILES_SECTION_HEADING,
    VARIABLES_SECTION_HEADING,
    CODE_SECTION_HEADING
)


def create_inline_prompt(
    prefix: str,
    suffix: str,
    variables: List[str],
    files: List[str]
) -> str:
    variables_str = '\n'.join([f"{variable}" for variable in variables])
    files_str = '\n'.join([f"file_name: {file}" for file in files])
    
    prompt = f"""You are a coding assistant that lives inside of JupyterLab. Your job is to help the user write code.

You're given the current code cell, the user's cursor position, and the variables defined in the notebook. The user's cursor is signified by the symbol <cursor>.

CRITICAL FORMATTING RULES:
1. Include a new line character at the start of your response if you want the code you are writing to be added on the line after the cursor. For example, if the cursor is at the end of a comment, you should start your response with a newline character so that the code you write is not added to the comment.
2. If you are finishing a line of code that the user started, return the full line of code with no newline character at the start or end.
3. Your response must preserve correct Python indentation and spacing. For example, if you're completing a line of indented code, you must preserve the indentation.

Your job is to complete the code that matches the user's intent. Write the minimal code to achieve the user's intent. Don't expand upon the user's intent.

<Example 1>
{FILES_SECTION_HEADING}
file_name: sales.csv

{VARIABLES_SECTION_HEADING} 
{{
    'loan_multiplier': 1.5,
    'sales_df': pd.DataFrame({{
        'transaction_date': ['2024-01-02', '2024-01-02', '2024-01-02', '2024-01-02', '2024-01-03'],
        'price_per_unit': [10, 9.99, 13.99, 21.00, 100],
        'units_sold': [1, 2, 1, 4, 5],
        'total_price': [10, 19.98, 13.99, 84.00, 500]
    }})
}}

{CODE_SECTION_HEADING}
```python
import pandas as pd
sales_df = pd.read_csv('./sales.csv')

# Multiply the total_price column by the loan_multiplier<cursor>
```

Output:
```python

sales_df['total_price'] = sales_df['total_price'] * loan_multiplier
```
</Example 1>

IMPORTANT: Notice in Example 1 that the output starts with a newline because the cursor was at the end of a comment. This newline is REQUIRED to maintain proper Python formatting.

<Example 2>
{FILES_SECTION_HEADING}


{VARIABLES_SECTION_HEADING} 
{{
    df: pd.DataFrame({{
        'age': [20, 25, 22, 23, 29],
        'name': ['Nawaz', 'Aaron', 'Charlie', 'Tamir', 'Eve'],
    }})
}}

{CODE_SECTION_HEADING}
```python
df['age'] = df[<cursor>['age'] > 23]
```

Output:
```python
df['age'] = df[df['age'] > 23]
```
</Example 2>

IMPORTANT: Notice in Example 2 that the output does NOT start with a newline because the cursor is in the middle of existing code.

<Example 3>
{FILES_SECTION_HEADING}
file_name: voters.csv

{VARIABLES_SECTION_HEADING} 
{{}}

{CODE_SECTION_HEADING}
```python
voters = pd.read_csv('./voters.csv')

# Create a variable for pennsylvania voters, ohio voters, california voters, and texas voters
pa_voters = voters[voters['state'] == 'PA']
ohio_voters<cursor>
```

Output:
```python
ohio_voters = voters[voters['state'] == 'OH']
ca_voters = voters[voters['state'] == 'CA']
tx_voters = voters[voters['state'] == 'TX']
```
</Example 3>

IMPORTANT: Notice in Example 3 that output does not start with a newline character because it wasnts to continue the line of code that the user started. Also notice the output contains three lines of code because that is the minimal code to achieve the user's intent.

<Example 4>
{FILES_SECTION_HEADING}
file_name: july_2025.xlsx
file_name: august_2025.xlsx

{VARIABLES_SECTION_HEADING} 
{{}}

{CODE_SECTION_HEADING}
```python
# Display the first 5 rows of the dataframe
df.head()
<cursor>
```

Output:
```python
```
</Example 4>

IMPORTANT: Notice in Example 4 that the output is empty becuase the user's intent is already complete.

<Example 5>
{FILES_SECTION_HEADING}


{VARIABLES_SECTION_HEADING} 
{{}}

{CODE_SECTION_HEADING}
```python
def even_and_odd():
    for i in range(10):
        if i % 2 == 0:
            print(f"Even: {{i}}")
        else:
            pri<cursor>
```

Output:
```python
            print(f"Odd: {{i}}")
```
</Example 5>

IMPORTANT: Notice in Example 5 that the output is indented several times because the code must be executed as part of the else block.

<Example 6>
{FILES_SECTION_HEADING}


{VARIABLES_SECTION_HEADING} 
{{}}

{CODE_SECTION_HEADING}
```python
days_in_week <cursor>
```

Output:
```python
days_in_week = 7
```
</Example 6>

IMPORTANT: Notice in Example 6 that inorder to finish the variable declaration, the output continues the existing line of code and does not start with a new line character.

Your Task:

{FILES_SECTION_HEADING}
{files_str}

{VARIABLES_SECTION_HEADING} 
{variables_str}

{CODE_SECTION_HEADING}
```python
{prefix}<cursor>{suffix}
```

Output:
"""
    return prompt


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

"""
Version check endpoint for Mito AI to compare local and PyPI versions.
"""
import json
import logging
import time
from functools import lru_cache
from typing import Tuple, Optional, Any
import requests
import tornado.web
from tornado.web import RequestHandler
from mito_ai._version import __version__

logger = logging.getLogger(__name__)

class VersionCheckHandler(RequestHandler):
    """Handler for checking the latest version of Mito AI on PyPI."""
    
    # Cache PyPI results for 1 hour (3600 seconds)
    @staticmethod
    @lru_cache(maxsize=1)
    def _get_latest_version() -> Tuple[Optional[str], float]:
        """Get the latest version from PyPI with caching."""
        cache_time = time.time()
        try:
            response = requests.get("https://pypi.org/pypi/mito-ai/json", timeout=3)
            response.raise_for_status()
            pypi_data = response.json()
            latest_version = pypi_data.get("info", {}).get("version", "")
            return latest_version, cache_time
        except (requests.RequestException, json.JSONDecodeError) as e:
            logger.warning(f"Failed to fetch latest version from PyPI: {str(e)}")
            return None, cache_time
    
    def get(self) -> None:
        """Get the current and latest Mito AI versions."""
        try:
            # Get current package version from _version.py
            current_version = __version__
            
            # Get latest version from PyPI with caching
            latest_version, cache_time = self._get_latest_version()
            cache_age = time.time() - cache_time
            
            # Include cache information in response
            self.set_header("Content-Type", "application/json")
            self.write(json.dumps({
                "current_version": current_version,
                "latest_version": latest_version,
                "cache_age_seconds": int(cache_age)
            }))
        except Exception as e:
            logger.error(f"Error in version check: {str(e)}")
            self.set_status(500)
            self.write(json.dumps({
                "error": "Failed to check versions"
            })) 


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import re
from typing import List
from openai.types.chat import ChatCompletionMessageParam
from mito_ai.completions.prompt_builders.prompt_constants import (
    ACTIVE_CELL_ID_SECTION_HEADING,
    ACTIVE_CELL_OUTPUT_SECTION_HEADING,
    GET_CELL_OUTPUT_TOOL_RESPONSE_SECTION_HEADING,
    FILES_SECTION_HEADING,
    VARIABLES_SECTION_HEADING,
    JUPYTER_NOTEBOOK_SECTION_HEADING,
    CONTENT_REMOVED_PLACEHOLDER
)


def trim_sections_from_message_content(content: str) -> str:
    """
    Removes specific metadata sections from message content to reduce token count so
    that users don't exceed the token limit for the LLM.

    These sections are replaced with a placeholder text.
    """

    # Replace metadata sections with placeholders
    section_headings = [
        FILES_SECTION_HEADING,
        VARIABLES_SECTION_HEADING,
        JUPYTER_NOTEBOOK_SECTION_HEADING,
        GET_CELL_OUTPUT_TOOL_RESPONSE_SECTION_HEADING,
        ACTIVE_CELL_OUTPUT_SECTION_HEADING,
        ACTIVE_CELL_ID_SECTION_HEADING
    ]
    
    for heading in section_headings:
        content = re.sub(
            f"{re.escape(heading)}\n(?:.+\n)+",
            f"{heading} {CONTENT_REMOVED_PLACEHOLDER}\n",
            content,
        )

    return content


def trim_old_messages(messages: List[ChatCompletionMessageParam], keep_recent: int = 3) -> List[ChatCompletionMessageParam]:
    """
    Trims metadata sections from messages that are older than the specified number of recent messages.
    We do this in order to reduce the token count of the messages, which helps us stay under the token limit for the LLM.
    """
    if len(messages) <= keep_recent:
        return messages
        
    # Process all messages except the keep_recent most recent ones. 
    # Only trim user messages, which is where this metadata lives. 
    # We want to not edit the system messages, as they contain important information / examples.
    for i in range(len(messages) - keep_recent):
        content = messages[i].get("content")
        
        is_user_message = messages[i].get("role") == "user"
        if not is_user_message: 
            continue
        
        content = messages[i].get("content")
        
        if content is None:
            continue
        
        if isinstance(content, str):
            # If content is just a string, then we just trim the metadata sections
            messages[i]["content"] = trim_sections_from_message_content(content)
        else: 
            # Otherwise, we get rid of the image_url section and just keep the trimmed text
            # We assume that there is only one text section in the content
            text_content = ""
            for section in content:
                if section.get("type") == "text" and "text" in section:
                    text_content = section["text"] #type: ignore
                    break
                
            messages[i]["content"] = trim_sections_from_message_content(text_content)        


    return messages


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import os
import pytest
import shutil
from mito_ai.rules.utils import RULES_DIR_PATH

# Define backup path for the rules directory
BACKUP_RULES_DIR_PATH = RULES_DIR_PATH + "_backup"


@pytest.fixture(scope="module", autouse=True)
def backup_rules_directory():
    """Backup the rules directory before tests and restore it after."""
    if os.path.exists(RULES_DIR_PATH):
        # Create backup, so we can test with a clean rules directory
        shutil.move(RULES_DIR_PATH, BACKUP_RULES_DIR_PATH)

    yield

    # Cleanup after tests
    if os.path.exists(RULES_DIR_PATH):
        shutil.rmtree(RULES_DIR_PATH)
    if os.path.exists(BACKUP_RULES_DIR_PATH):
        shutil.move(BACKUP_RULES_DIR_PATH, RULES_DIR_PATH)



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import List, Tuple, Any
from jupyter_server.utils import url_path_join
from mito_ai.db.handlers import ConnectionsHandler, SchemaHandler


def get_db_urls(base_url: str) -> List[Tuple[str, Any, dict]]:
    """Get all database related URL patterns.

    Args:
        base_url: The base URL for the Jupyter server

    Returns:
        List of (url_pattern, handler_class, handler_kwargs) tuples
    """
    BASE_URL = base_url + "/mito-ai/db"
    return [
        (url_path_join(BASE_URL, "connections"), ConnectionsHandler, {}),
        (
            # URL for deleting a connection.
            url_path_join(BASE_URL, "connections", "(?P<uuid>[^/]+)"),
            ConnectionsHandler,
            {},
        ),
        (url_path_join(BASE_URL, "schemas"), SchemaHandler, {}),
        (
            # URL for deleting a schema.
            url_path_join(BASE_URL, "schemas", "(?P<uuid>[^/]+)"),
            SchemaHandler,
            {},
        ),
    ]



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import requests

RULE_NAME = "test_rule"
RULE_CONTENT = "This is a test rule for data analysis."

from mito_ai.tests.conftest import TOKEN

# --- PUT RULES ---


def test_put_rule_with_auth(jp_base_url):
    response = requests.put(
        jp_base_url + f"/mito-ai/rules/{RULE_NAME}",
        headers={"Authorization": f"token {TOKEN}"},
        json={"content": RULE_CONTENT},
    )
    assert response.status_code == 200

    response_json = response.json()
    assert response_json["status"] == "updated"
    assert response_json["rules file "] == RULE_NAME


def test_put_rule_with_no_auth(jp_base_url):
    response = requests.put(
        jp_base_url + f"/mito-ai/rules/{RULE_NAME}",
        json={"content": RULE_CONTENT},
    )

    assert response.status_code == 403  # Forbidden


def test_put_rule_with_incorrect_auth(jp_base_url):
    response = requests.put(
        jp_base_url + f"/mito-ai/rules/{RULE_NAME}",
        headers={"Authorization": f"token incorrect-token"}, # <- wrong token
        json={"content": RULE_CONTENT},
    )
    assert response.status_code == 403  # Forbidden


def test_put_rule_missing_content(jp_base_url):
    response = requests.put(
        jp_base_url + f"/mito-ai/rules/{RULE_NAME}",
        headers={"Authorization": f"token {TOKEN}"},
        json={},  # Missing content field
    )
    assert response.status_code == 400  # Bad Request


# --- GET SPECIFIC RULE ---


def test_get_rule_with_auth(jp_base_url):
    response = requests.get(
        jp_base_url + f"/mito-ai/rules/{RULE_NAME}",
        headers={"Authorization": f"token {TOKEN}"},
    )
    assert response.status_code == 200
    assert response.json() == {"key": RULE_NAME, "content": RULE_CONTENT}


def test_get_rule_with_no_auth(jp_base_url):
    response = requests.get(
        jp_base_url + f"/mito-ai/rules/{RULE_NAME}",
    )
    assert response.status_code == 403  # Forbidden


def test_get_rule_with_incorrect_auth(jp_base_url):
    response = requests.get(
        jp_base_url + f"/mito-ai/rules/{RULE_NAME}",
        headers={"Authorization": f"token incorrect-token"}, # <- wrong token
    )
    assert response.status_code == 403  # Forbidden


def test_get_nonexistent_rule_with_auth(jp_base_url):
    response = requests.get(
        jp_base_url + f"/mito-ai/rules/nonexistent_rule",
        headers={"Authorization": f"token {TOKEN}"},
    )
    assert response.status_code == 404  # Not Found


# --- GET ALL RULES ---


def test_get_all_rules_with_auth(jp_base_url):
    response = requests.get(
        jp_base_url + f"/mito-ai/rules",
        headers={"Authorization": f"token {TOKEN}"},
    )
    assert response.status_code == 200
    
    response_json = response.json()
    assert isinstance(response_json, list)
    # Should contain our test rule (with .md extension)
    assert f"{RULE_NAME}.md" in response_json


def test_get_all_rules_with_no_auth(jp_base_url):
    response = requests.get(
        jp_base_url + f"/mito-ai/rules",
    )
    assert response.status_code == 403  # Forbidden


def test_get_all_rules_with_incorrect_auth(jp_base_url):
    response = requests.get(
        jp_base_url + f"/mito-ai/rules",
        headers={"Authorization": f"token incorrect-token"},
    )
    assert response.status_code == 403  # Forbidden



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import os
from typing import Final
from mito_ai.utils.schema import MITO_FOLDER

DB_DIR_PATH: Final[str] = os.path.join(MITO_FOLDER, "db")
CONNECTIONS_PATH: Final[str] = os.path.join(DB_DIR_PATH, "connections.json")
SCHEMAS_PATH: Final[str] = os.path.join(DB_DIR_PATH, "schemas.json")
BACKUP_DB_DIR_PATH: Final[str] = os.path.join(MITO_FOLDER, "db_backup")
MSSQL_CONNECTION_DETAILS = {
    "type": "mssql",
    "username": os.environ.get("MSSQL_USERNAME"),
    "password": os.environ.get("MSSQL_PASSWORD"),
    "host": os.environ.get("MSSQL_HOST"),
    "port": "1433",
    "database": "Northwind",
    "odbc_driver_version": "18",
}
MYSQL_CONNECTION_DETAILS = {
    "type": "mysql",
    "username": os.environ.get("MYSQL_USERNAME"),
    "password": os.environ.get("MYSQL_PASSWORD"),
    "host": os.environ.get("MYSQL_HOST"),
    "port": "3306",
    "database": "Northwind",
}
ORACLE_CONNECTION_DETAILS = {
    "type": "oracle",
    "username": os.environ.get("ORACLE_USERNAME"),
    "password": os.environ.get("ORACLE_PASSWORD"),
    "host": os.environ.get("ORACLE_HOST"),
    "port": "1521",
    "service_name": "ORCL",
}
POSTGRES_CONNECTION_DETAILS = {
    "type": "postgres",
    "username": os.environ.get("POSTGRES_USERNAME"),
    "password": os.environ.get("POSTGRES_PASSWORD"),
    "host": os.environ.get("POSTGRES_HOST"),
    "port": "5432",
    "database": "postgres",
}
SNOWFLAKE = {
    "type": "snowflake",
    "username": os.environ.get("SNOWFLAKE_USERNAME"),
    "password": os.environ.get("SNOWFLAKE_PASSWORD"),
    "account": os.environ.get("SNOWFLAKE_ACCOUNT"),
    "warehouse": "COMPUTE_WH",
}
SQLITE_TEST_DB_PATH: Final[str] = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
    "tests",
    "data",
    "stock_data.sqlite3",
)
SQLITE_CONNECTION_DETAILS = {
    "type": "sqlite",
    "database": SQLITE_TEST_DB_PATH,
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
Utilities for creating and initializing the user.json
file with the current schema
"""

import json
import os
from typing import Literal, Optional

from mito_ai.utils.db import (MITO_FOLDER, USER_JSON_PATH, set_user_field)
from mito_ai.utils.schema import (GITHUB_ACTION_EMAIL, GITHUB_ACTION_ID,
                                    UJ_STATIC_USER_ID, UJ_USER_EMAIL,
                                    USER_JSON_DEFAULT)
from mito_ai.utils.telemetry_utils import identify
from mito_ai.utils.utils import is_running_test


def is_user_json_exists_and_valid_json() -> bool:
    """
    Helper function that determines if the current user.json both
    exists and is valid json
    """
    if not os.path.exists(USER_JSON_PATH):
        return False

    try:
        with open(USER_JSON_PATH, 'r') as f:
            json.loads(f.read())
        return True
    except:
        return False


def try_create_user_json_file() -> None:
    
    # Create the mito folder if it does not exist
    if not os.path.exists(MITO_FOLDER):
        os.mkdir(MITO_FOLDER)

    # We create a user.json file if it does not exist, or if it
    # is invalid (e.g. it is not parseable JSON).
    if not is_user_json_exists_and_valid_json():
        # First, we write an empty default object
        with open(USER_JSON_PATH, 'w+') as f:
            f.write(json.dumps(USER_JSON_DEFAULT))

        # Then, we take special care to put all the testing/CI environments 
        # (e.g. Github actions) under one ID and email
        if is_running_test():
            set_user_field(UJ_STATIC_USER_ID, GITHUB_ACTION_ID)
            set_user_field(UJ_USER_EMAIL, GITHUB_ACTION_EMAIL)


def initialize_user() -> None:
    """
    Internal helper function that gets called whenever a ai completion is requested.

    It:
    1. Creates the user.json if it does not exist (though it usually does, from the installer)
    2. Identifies the user
    """
    # Try to create the user.json file, if it does not already exist
    try_create_user_json_file()

    # Identify the user 
    identify()

    # Note, its possible that a user has a previous version of the user.json file if they 
    # downloaded Mito >1 year ago and have not yet upgraded. In this case, we would like to 
    # upgrade the user.json to the newest versio. But for simplicity, let's not 
    # try to upgrade it here. We will either move away from user.json all together
    # or unify the utilities for mitosheet and mitoai.


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the The Mito Enterprise license.

from mito_ai.utils.version_utils import is_enterprise, is_mitosheet_private
from mito_ai.constants import AZURE_OPENAI_API_KEY, AZURE_OPENAI_ENDPOINT, AZURE_OPENAI_API_VERSION, AZURE_OPENAI_MODEL

def is_azure_openai_configured() -> bool:
    """
    Azure OpenAI is only supported for Mito Enterprise users
    """
    is_allowed_to_use_azure = is_enterprise() or is_mitosheet_private()
    return all([is_allowed_to_use_azure, AZURE_OPENAI_API_KEY, AZURE_OPENAI_ENDPOINT, AZURE_OPENAI_API_VERSION, AZURE_OPENAI_MODEL])


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import os
from typing import cast
from .schema import UJ_MITOSHEET_ENTERPRISE, UJ_MITOSHEET_PRO
from .db import get_user_field

# Check if helper packages are installed
try:
    import mitosheet_helper_pro
    MITOSHEET_HELPER_PRO = True
except ImportError:
    MITOSHEET_HELPER_PRO = False
try:
    import mitosheet_helper_enterprise  
    MITOSHEET_HELPER_ENTERPRISE = True
except ImportError:
    MITOSHEET_HELPER_ENTERPRISE = False

try:
    import mitosheet_private 
    MITOSHEET_PRIVATE = True
except ImportError:
    MITOSHEET_PRIVATE = False

# This is a legacy helper that we don't use anymore, however, we're keeping it for now
# for backwards compatibility, since I'm not 100% confident that nobody is currently using it.
try:
    import mitosheet_helper_private 
    MITOSHEET_HELPER_PRIVATE = True
except ImportError:
    MITOSHEET_HELPER_PRIVATE = False


def is_pro() -> bool:
    """
    Helper function for returning if this is a
    pro deployment of mito
    """

    # This package overides the user.json
    if MITOSHEET_HELPER_PRO:
        return True

    # This package overides the user.json
    if MITOSHEET_PRIVATE:
        return True

    # Check if the config is set
    # TODO: Check if the mito config pro is set to true.
    # I don't think that any user is on pro via this method

    # If you're on Mito Enterprise, then you get all Mito Pro features
    if is_enterprise():
        return True

    pro = get_user_field(UJ_MITOSHEET_PRO)
    if pro is None:
        return False
    
    return bool(pro)

def is_enterprise() -> bool:
    """
    Helper function for returning if this is a Mito Enterprise
    users
    """

    # This package overides the user.json
    if MITOSHEET_HELPER_ENTERPRISE:
        return True
    
    # TODO: Check if the mito config enterprise is set to true. 
    # I don't think that any user is on enterprise via this method
    
    is_enterprise = get_user_field(UJ_MITOSHEET_ENTERPRISE)
    if is_enterprise is None:
        return False

    # TODO: Check if someone has a temp enterprise license set

    return bool(is_enterprise)


def is_mitosheet_private() -> bool:
    """
    Helper function for returning if this is a Mito Private
    users
    """

    if MITOSHEET_PRIVATE:
        return True
    return False



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import List
from mito_ai.completions.prompt_builders.prompt_constants import (
    ACTIVE_CELL_ID_SECTION_HEADING,
    FILES_SECTION_HEADING,
    VARIABLES_SECTION_HEADING,
    CODE_SECTION_HEADING
)


def create_error_prompt(
    error_message: str,
    active_cell_code: str,
    active_cell_id: str,
    variables: List[str],
    files: List[str]
) -> str:
    variables_str = '\n'.join([f"{variable}" for variable in variables])
    files_str = '\n'.join([f"{file}" for file in files])
    return f"""Help me debug this code in JupyterLab. Analyze the error and provide a solution that maintains the original intent.

<Example 1>
{FILES_SECTION_HEADING}
file_name: sales.csv

{VARIABLES_SECTION_HEADING}
{{
    'revenue_multiplier': 1.5,
    'sales_df': pd.DataFrame({{
        'transaction_date': ['2024-01-02', '2024-01-02', '2024-01-02', '2024-01-02', '2024-01-03'],
        'price_per_unit': [10, 9.99, 13.99, 21.00, 100],
        'units_sold': [1, 2, 1, 4, 5],
        'total_price': [10, 19.98, 13.99, 84.00, 500]
    }})
}}

{ACTIVE_CELL_ID_SECTION_HEADING}
'9e38c62b-38f8-457d-bb8d-28bfc52edf2c'

{CODE_SECTION_HEADING}
```python
import pandas as pd
sales_df = pd.read_csv('./sales.csv')
revenue_multiplier =  1.5
sales_df['total_revenue'] = sales_df['price'] * revenue_multiplier
```

Error Traceback:
Cell In[24], line 4
      1 import pandas as pd
      2 sales_df = pd.read_csv('./sales.csv')
      3 revenue_multiplier =  1.5
----> 4 sales_df['total_revenue'] = sales_df['price'] * revenue_multiplier

KeyError: 'price'


ERROR ANALYSIS:
Runtime error: Attempted to access non-existent DataFrame column

INTENT ANALYSIS:
User is trying to calculate total revenue by applying a multiplier to transaction prices. Based on the defined variables, the column that the user is tring to access is likely `total_price` because that would allow them to calculate the total revenue for each transaction.

SOLUTION:
```python
import pandas as pd
sales_df = pd.read_csv('./sales.csv')
revenue_multiplier =  1.5
sales_df['total_revenue'] = sales_df['total_price'] * revenue_multiplier
```

The DataFrame contains 'total_price' rather than 'price'. Updated column reference to match existing data structure.
</Example 1>

<Example 2>
{FILES_SECTION_HEADING}


{VARIABLES_SECTION_HEADING}
{{
    'df': pd.DataFrame({{
        'order_id': [1, 2, 3, 4],
        'date': ['Mar 7, 2025', 'Sep 24, 2024', '25 June, 2024', 'June 29, 2024'],
        'amount': [100, 150, 299, 99]
    }})
}}

{ACTIVE_CELL_ID_SECTION_HEADING}
'c68fdf19-db8c-46dd-926f-d90ad35bb3bc'

{CODE_SECTION_HEADING}
```python
df['date'] = pd.to_datetime(df['date'])
```

Error Traceback:
Cell In[27], line 1
----> 1 df['date'] = pd.to_datetime(df['date'])

ValueError: time data "25 June, 2024" doesn't match format "%b %d, %Y", at position 2. You might want to try:
    - passing `format` if your strings have a consistent format;
    - passing `format='ISO8601'` if your strings are all ISO8601 but not necessarily in exactly the same format;
    - passing `format='mixed'`, and the format will be inferred for each element individually. You might want to use `dayfirst` alongside this.

ERROR ANALYSIS:
This is a ValueError caused by applying the wrong format to a specific date string. Because it was triggered at position 2, the first date string must have successfully converted. By looking at the defined variables, I can see that first date string is in the format "Mar 7, 2025", but the third date string is in the format "25 June, 2024". Those dates are not in the same format, so the conversion failed.

INTENT ANALYSIS:
User is trying to convert the date column to a datetime object even though the dates are not in the same starting format. 

SOLUTION:
```python
def parse_date(date_str):
    formats = ['%b %d, %Y', '%d %B, %Y']
    
    for fmt in formats:
        try:
            return pd.to_datetime(date_str, format=fmt)
        except ValueError:
            # Try the next format
            continue
            
    # If no format worked, return Not a Time
    return pd.NaT

df['date'] = df['date'].apply(lambda x: parse_date(x))
```

Since the dates are not in a consistent format, we need to first figure out which format to use for each date string and then use that format to convert the date.

The best way to do this is with a function. We can call this function `parse_date`.
</Example 2>


Guidelines for Solutions:

Error Analysis:

- Identify error type (Syntax, Runtime, Logic).
- Use the defined variables and code in the active cell to understand the error.
- Consider kernel state and execution order

Intent Preservation:

- Try to understand the user's intent using the defined variables and code in the active cell.

Solution Requirements:

- Return the full code cell with the error fixed and a short explanation of the error.
- Only update code in the active cell. Do not update other code in the notebook.
- Propose a solution that fixes the error and does not change the user's intent.
- Make the solution as simple as possible.
- Reuse as much of the existing code as possible.
- Do not add temporary comments like '# Fixed the typo here' or '# Added this line to fix the error'
- The code in the SOLUTION section should be a python code block starting with ```python and ending with ```
- If you encounter a ModuleNotFoundError, you can install the package by adding the the following line to the top of the code cell: `!pip install <package_name> --quiet`.

Here is your task. 

{FILES_SECTION_HEADING}
{files_str}

{VARIABLES_SECTION_HEADING}
{variables_str}

{ACTIVE_CELL_ID_SECTION_HEADING}
{active_cell_id}

{CODE_SECTION_HEADING}
```python
{active_cell_code}
```

Error Traceback:
{error_message}

ERROR ANALYSIS:

INTENT ANALYSIS:

SOLUTION:
"""


def remove_inner_thoughts_from_message(message: str) -> str:
    # The smart debug prompt thinks to itself before returning the solution. We don't need to save the inner thoughts. 
    # We remove them before saving the message in the chat history
    if message == "":
        return message
    
    SOLUTION_STRING = "SOLUTION:"

    if SOLUTION_STRING in message:
        message = message.split(SOLUTION_STRING)[1].strip()
    
    return message




# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import Final
from mito_ai.utils.db import (
    get_chat_completion_count, 
    get_autocomplete_count,
    get_first_completion_date, 
    get_last_reset_date,
    get_user_field, 
    set_user_field
)
from mito_ai.utils.schema import (
    UJ_MITO_AI_FIRST_USAGE_DATE, 
    UJ_AI_MITO_API_NUM_USAGES, 
    UJ_MITO_AI_LAST_RESET_DATE,
    UJ_AI_MITO_AUTOCOMPLETE_NUM_USAGES
)
from mito_ai.utils.telemetry_utils import (MITO_SERVER_FREE_TIER_LIMIT_REACHED, log)
from mito_ai.utils.version_utils import is_pro
from mito_ai.completions.models import MessageType
from datetime import datetime, timedelta

"""
IT IS EXPRESSLY AGAINST THE SOFTWARE LICENSE TO MODIFY, BYPASS, OR OTHERWISE
CIRCUMVENT THE QUOTA LIMITS OF THE FREE TIER. We want to provide users with a 
free tier, but running AI models is expensive, so we need to limit the usage 
or we will no longer be able to provide this free tier.
"""
# Monthly chat completions limit for free tier users
OS_MONTHLY_AI_COMPLETIONS_LIMIT: Final[int] = 50

# Monthly autocomplete limit for free tier users
OS_MONTHLY_AUTOCOMPLETE_LIMIT: Final[int] = 5000

def check_mito_server_quota(message_type: MessageType) -> None:
    """
    Checks if the user has exceeded their Mito server quota. Pro users have no limits.
    Raises PermissionError if the user has exceeded their quota.
    
    IT IS EXPRESSLY AGAINST THE SOFTWARE LICENSE TO MODIFY, BYPASS, OR OTHERWISE
    CIRCUMVENT THE QUOTA LIMITS OF THE FREE TIER. We want to provide users with a 
    free tier, but running AI models is expensive, so we need to limit the usage 
    or we will no longer be able to provide this free tier.
    """
    if is_pro():
        return

    # Get current date
    current_date = datetime.now()
    
    # Get the date when the last reset occurred
    last_reset_date_str = get_last_reset_date()
        
    # If no last reset date is found, set it to today and reset both counters
    if last_reset_date_str is None:
        set_user_field(UJ_MITO_AI_LAST_RESET_DATE, current_date.strftime("%Y-%m-%d"))
        set_user_field(UJ_AI_MITO_API_NUM_USAGES, 0)
        set_user_field(UJ_AI_MITO_AUTOCOMPLETE_NUM_USAGES, 0)
        return
    
    # Convert the last reset date string to a datetime object
    last_reset_date = datetime.strptime(last_reset_date_str, "%Y-%m-%d")
    
    # Calculate the date one month from the last reset
    one_month_later = last_reset_date + timedelta(days=30)  # Approximating a month as 30 days
    
    # If it's been more than a month since the last reset, reset both counters
    if current_date >= one_month_later:
        set_user_field(UJ_AI_MITO_API_NUM_USAGES, 0)
        set_user_field(UJ_AI_MITO_AUTOCOMPLETE_NUM_USAGES, 0)
        set_user_field(UJ_MITO_AI_LAST_RESET_DATE, current_date.strftime("%Y-%m-%d"))
        
    # Check the appropriate limit based on message type
    if message_type == MessageType.INLINE_COMPLETION:
        # Check autocomplete limit
        autocomplete_count = get_autocomplete_count()
        
        if autocomplete_count >= OS_MONTHLY_AUTOCOMPLETE_LIMIT:
            log(MITO_SERVER_FREE_TIER_LIMIT_REACHED)
            raise PermissionError(MITO_SERVER_FREE_TIER_LIMIT_REACHED)
    else:
        # Check chat completion limit
        completion_count = get_chat_completion_count()
        
        if completion_count >= OS_MONTHLY_AI_COMPLETIONS_LIMIT:
            log(MITO_SERVER_FREE_TIER_LIMIT_REACHED)
            raise PermissionError(MITO_SERVER_FREE_TIER_LIMIT_REACHED)
        

def update_mito_server_quota(message_type: MessageType) -> None:
    """
    Update the user's quota for the Mito Server.
    
    IT IS EXPRESSLY AGAINST THE SOFTWARE LICENSE TO MODIFY, BYPASS, OR OTHERWISE
    CIRCUMVENT THE QUOTA LIMITS OF THE FREE TIER. We want to provide users with a 
    free tier, but running AI models is expensive, so we need to limit the usage 
    or we will no longer be able to provide this free tier.
    """
    
    if message_type == MessageType.CHAT_NAME_GENERATION:
        # We do not count the Chat Name Generation message type towards the quota
        return
    
    current_date = datetime.now()
    first_usage_date = get_first_completion_date()
    last_reset_date_str = get_last_reset_date()
    
    if first_usage_date is None:
        first_usage_date = current_date.strftime("%Y-%m-%d")
        set_user_field(UJ_MITO_AI_FIRST_USAGE_DATE, first_usage_date)
    
    # Initialize the reset date if it doesn't exist
    if last_reset_date_str is None:
        last_reset_date_str = current_date.strftime("%Y-%m-%d")
        set_user_field(UJ_MITO_AI_LAST_RESET_DATE, last_reset_date_str)
        
    # Convert the last reset date string to a datetime object
    last_reset_date = datetime.strptime(last_reset_date_str, "%Y-%m-%d")
    
    # Calculate the date one month from the last reset
    one_month_later = last_reset_date + timedelta(days=30)
    
    # If it's been more than a month since the last reset, reset both counters
    # and make today's date the new reset date
    if current_date >= one_month_later:
        set_user_field(UJ_AI_MITO_API_NUM_USAGES, 0)
        set_user_field(UJ_AI_MITO_AUTOCOMPLETE_NUM_USAGES, 0)
        last_reset_date_str = current_date.strftime("%Y-%m-%d")
        set_user_field(UJ_MITO_AI_LAST_RESET_DATE, last_reset_date_str)
    
    # Update the appropriate usage counter based on message type
    if message_type == MessageType.INLINE_COMPLETION:
        # Increment autocomplete count
        autocomplete_count = get_autocomplete_count()
        autocomplete_count = 1 if autocomplete_count is None else autocomplete_count + 1
        
        try:
            set_user_field(UJ_AI_MITO_AUTOCOMPLETE_NUM_USAGES, autocomplete_count)
        except Exception as e:
            raise e
    else:
        # Increment chat completion count
        completion_count = get_chat_completion_count()
        completion_count = 1 if completion_count is None else completion_count + 1
        
        try:
            set_user_field(UJ_AI_MITO_API_NUM_USAGES, completion_count)
            set_user_field(UJ_MITO_AI_FIRST_USAGE_DATE, first_usage_date)
        except Exception as e:
            raise e
        


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import os
import json
import pytest
import shutil
from typing import Generator
from mito_ai.tests.db.test_db_constants import (
    DB_DIR_PATH,
    CONNECTIONS_PATH,
    BACKUP_DB_DIR_PATH,
)


@pytest.fixture(scope="module", autouse=True)
def backup_db_folder() -> Generator[None, None, None]:
    """Backup the DB folder before tests and restore it after."""
    if os.path.exists(DB_DIR_PATH):
        # Create backup, so we can test with a clean db dir
        shutil.move(DB_DIR_PATH, BACKUP_DB_DIR_PATH)

    yield

    # Cleanup after tests
    if os.path.exists(DB_DIR_PATH):
        shutil.rmtree(DB_DIR_PATH)
    if os.path.exists(BACKUP_DB_DIR_PATH):
        shutil.move(BACKUP_DB_DIR_PATH, DB_DIR_PATH)


@pytest.fixture
def first_connection_id() -> str:
    # Manually open the connections.json file and get the first connection ID
    with open(CONNECTIONS_PATH, "r") as f:
        connections = json.load(f)
    # Get the first connection ID from the object keys
    connection_id = next(iter(connections.keys()))
    return str(connection_id)



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import requests
from mito_ai.tests.db.test_db_constants import POSTGRES_CONNECTION_DETAILS
from mito_ai.tests.conftest import TOKEN

# To create a postgres database, run the following command:
# docker-compose -f mito_ai/docker/postgres/compose.yml up
# and then, to delete the database, run the following command:
# docker-compose -f mito_ai/docker/postgres/compose.yml down -v


def test_add_postgres_connection(jp_base_url: str) -> None:
    response = requests.post(
        jp_base_url + "/mito-ai/db/connections",
        headers={"Authorization": f"token {TOKEN}"},
        json=POSTGRES_CONNECTION_DETAILS,
    )
    assert response.status_code == 200

    # Get the schemas
    response = requests.get(
        jp_base_url + f"/mito-ai/db/schemas",
        headers={"Authorization": f"token {TOKEN}"},
    )
    assert response.status_code == 200
    # Esnure that there is one scema dict in the response
    assert len(response.json()) == 1



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import pytest
from mito_ai.completions.prompt_builders.agent_system_message import create_agent_system_message_prompt

def test_create_agent_system_message_prompt_browser_conditional() -> None:
    """
    Test that verifies the create_agent_system_message_prompt function produces different output
    based on the browser type.
    
    Specifically:
    - When isChromeBrowser=False, the prompt should NOT contain GET_CELL_OUTPUT
    - When isChromeBrowser=True, the prompt SHOULD contain GET_CELL_OUTPUT
    """
    # Test non-Chrome browser case
    non_chrome_prompt = create_agent_system_message_prompt(isChromeBrowser=False)
    assert "GET_CELL_OUTPUT" not in non_chrome_prompt, "Non-Chrome prompt should not contain GET_CELL_OUTPUT"
    
    # Test Chrome browser case
    chrome_prompt = create_agent_system_message_prompt(isChromeBrowser=True)
    assert "GET_CELL_OUTPUT" in chrome_prompt, "Chrome prompt should contain GET_CELL_OUTPUT"


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from sqlalchemy import create_engine, text
from mito_ai.db.models import WarehouseDetails

SUPPORTED_DATABASE_KINDS = ["STANDARD", "IMPORTED DATABASE"]


def crawl_snowflake(username: str, password: str, account: str, warehouse: str) -> dict:
    try:
        conn_str = (
            f"snowflake://{username}:{password}@{account}/" f"?warehouse={warehouse}"
        )
        engine = create_engine(conn_str)

        # Step 1: Get databases
        db_query = text("SHOW DATABASES")
        with engine.connect() as connection:
            result = connection.execute(db_query)
            databases = result.mappings().all()

        # Step 2: Filter for 'STANDARD' and 'IMPORTED DATABASE' kinds
        filtered_db_names = [
            row["name"] for row in databases if row["kind"] in SUPPORTED_DATABASE_KINDS
        ]

        # Step 3: Escape and format for SQL IN clause
        db_list_sql = ", ".join(f"'{name}'" for name in filtered_db_names)

        # Step 4: Use filtered databases in schema query
        schema_query = text(
            f"""
            SELECT 
                c.table_catalog AS database_name,
                c.table_schema AS schema_name,
                c.table_name,
                c.column_name,
                c.data_type,
                c.comment
            FROM snowflake.account_usage.columns c
            WHERE c.deleted IS NULL
            AND c.table_catalog IN ({db_list_sql})
        """
        )

        # Step 5: Execute the schema query and process the results
        warehouse_details: WarehouseDetails = {"databases": {}}

        with engine.connect() as connection:
            result = connection.execute(schema_query)
            for row in result.fetchall():
                db = row[0]
                schema = row[1]
                table = row[2]
                column_name = row[3]

                db_dict = warehouse_details["databases"].setdefault(db, {"schemas": {}})
                schema_dict = db_dict["schemas"].setdefault(schema, {"tables": {}})
                columns_list = schema_dict["tables"].setdefault(table, [])
                columns_list.append({"name": column_name, "type": row[4]})

        return {
            "schema": warehouse_details,
            "error": None,
        }
    except Exception as e:
        return {
            "schema": None,
            "error": str(e),
        }



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import requests

SETTINGS_KEY = "super_secret_mode"
SETTINGS_VALUE = "yes"

from mito_ai.tests.conftest import TOKEN

# --- PUT SETTINGS ---


def test_put_settings_with_auth(jp_base_url: str) -> None:
    response = requests.put(
        jp_base_url + f"/mito-ai/settings/{SETTINGS_KEY}",
        headers={"Authorization": f"token {TOKEN}"},
        json={"value": SETTINGS_VALUE},
    )
    assert response.status_code == 200

    response_json = response.json()
    assert response_json["status"] == "updated"
    assert response_json["key"] == SETTINGS_KEY
    assert response_json["value"] == SETTINGS_VALUE


def test_put_settings_with_no_auth(jp_base_url: str) -> None:
    response = requests.put(
        jp_base_url + f"/mito-ai/settings/{SETTINGS_KEY}",
        json={"value": SETTINGS_VALUE},
    )

    assert response.status_code == 403  # Forbidden


def test_put_settings_with_incorrect_auth(jp_base_url: str) -> None:
    response = requests.put(
        jp_base_url + f"/mito-ai/settings/{SETTINGS_KEY}",
        headers={"Authorization": f"token incorrect-token"},
        json={"value": SETTINGS_VALUE},
    )
    assert response.status_code == 403  # Forbidden


# # --- GET SETTINGS ---


def test_get_settings_with_auth(jp_base_url: str) -> None:
    response = requests.get(
        jp_base_url + f"/mito-ai/settings/{SETTINGS_KEY}",
        headers={"Authorization": f"token {TOKEN}"},
    )
    assert response.status_code == 200
    assert response.json() == {"key": SETTINGS_KEY, "value": SETTINGS_VALUE}


def test_get_settings_with_no_auth(jp_base_url: str) -> None:
    response = requests.get(
        jp_base_url + f"/mito-ai/settings/{SETTINGS_KEY}",
    )
    assert response.status_code == 403  # Forbidden


def test_get_settings_with_incorrect_auth(jp_base_url: str) -> None:
    response = requests.get(
        jp_base_url + f"/mito-ai/settings/{SETTINGS_KEY}",
        headers={"Authorization": f"token incorrect-token"},
    )
    assert response.status_code == 403  # Forbidden



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import json
import anthropic
from typing import Dict, Any, Optional, Tuple, Union, Callable, List, cast

from anthropic.types import Message, MessageParam
from mito_ai.completions.models import ResponseFormatInfo, CompletionReply, CompletionStreamChunk, CompletionItem, MessageType
from openai.types.chat import ChatCompletionMessageParam
from mito_ai.utils.anthropic_utils import get_anthropic_completion_from_mito_server, stream_anthropic_completion_from_mito_server, get_anthropic_completion_function_params

# Max tokens is a required parameter for the Anthropic API. 
# We set it to a high number so that we can edit large code cells
# 8192 is the maximum allowed number of output tokens for claude-3-5-haiku-20241022
MAX_TOKENS = 8_000

ANTHROPIC_FAST_MODEL = "claude-3-5-haiku-latest"

def extract_and_parse_anthropic_json_response(response: Message) -> Union[object, Any]:
    """
    Extracts and parses the JSON response from the Claude API.
    """
    try:
        # Check for tool use in the response
        for content_block in response.content:
            if content_block.type == "tool_use" and content_block.name == "agent_response":
                result = content_block.input
                return result

        # If no tool use was found, try to parse the text response
        text_response = None
        for content_block in response.content:
            if content_block.type == "text":
                text_response = content_block.text
                break

        if text_response:
            # Try to extract JSON from the text response
            import re
            json_pattern = r'(\{.*\})'
            match = re.search(json_pattern, text_response, re.DOTALL)
            if match:
                try:
                    json_response = json.loads(match.group(0))
                    return json_response
                except json.JSONDecodeError:
                    pass

        raise Exception("No valid AgentResponse format found in the response")
    except Exception as e:
        raise Exception(f"Failed to parse response: {e}")


def get_anthropic_system_prompt_and_messages(messages: List[ChatCompletionMessageParam]) -> Tuple[
    Union[str, anthropic.NotGiven], List[MessageParam]]:
    """
    Convert a list of OpenAI messages to a list of Anthropic messages.
    """
    
    system_prompt: Union[str, anthropic.NotGiven] = anthropic.NotGiven()
    anthropic_messages: List[MessageParam] = []

    for message in messages:
        if 'content' not in message:
            continue

        # We assume that the conversation only has one system message.
        # Or if there are multiple, we take the last one.
        if message['role'] == 'system':
            system_prompt = str(message['content'])

        # Construct the messages for the user and assistant in Anthropic format.
        if message['role'] == 'user':
            content = message['content']

            # Handle mixed content (text + images)
            if isinstance(content, list):
                anthropic_content = []

                for item in content:
                    if isinstance(item, dict):
                        item_dict = cast(Dict[str, Any], item)
                        if item_dict.get('type') == 'text':
                            # Add text content
                            text_content = item_dict.get('text', '')
                            anthropic_content.append({
                                "type": "text",
                                "text": text_content
                            })
                        elif item_dict.get('type') == 'image_url':
                            # Convert OpenAI image format to Anthropic format
                            image_url_obj = item_dict.get('image_url', {})
                            if isinstance(image_url_obj, dict):
                                image_url = image_url_obj.get('url', '')
                            else:
                                image_url = str(image_url_obj)

                            # Extract media type and base64 data
                            if image_url.startswith('data:'):
                                # Format: data:image/png;base64,<base64_data>
                                header, base64_data = image_url.split(',', 1)
                                media_type = header.split(';')[0].split(':')[1]  # Extract image/png or image/jpeg
                            else:
                                # If it's not a data URL, assume it's direct base64 and default to image/png
                                media_type = "image/png"
                                base64_data = image_url

                            anthropic_content.append({
                                "type": "image",
                                "source": {
                                    "type": "base64",
                                    "media_type": media_type,
                                    "data": base64_data
                                }
                            })

                anthropic_messages.append(MessageParam(role='user', content=cast(Any, anthropic_content)))
            else:
                # Handle simple text content
                anthropic_messages.append(MessageParam(role='user', content=str(content)))

        elif message['role'] == 'assistant':
            anthropic_messages.append(MessageParam(role='assistant', content=str(message['content'])))

    return system_prompt, anthropic_messages


class AnthropicClient:
    """
    A client for interacting with the Anthropic API or the Mito server fallback.
    """

    def __init__(self, api_key: Optional[str], model: str, timeout: int = 30, max_retries: int = 1):
        self.api_key = api_key
        self.model = model
        self.timeout = timeout
        self.max_retries = max_retries
        self.client: Optional[anthropic.Anthropic]
        if api_key:
            self.client = anthropic.Anthropic(api_key=api_key)
        else:
            self.client = None

    async def request_completions(self, messages: List[ChatCompletionMessageParam],
                                  response_format_info: Optional[ResponseFormatInfo] = None,
                                  message_type: MessageType = MessageType.CHAT) -> Any:
        """
        Get a response from Claude or the Mito server that adheres to the AgentResponse format.
        """
        try:
            anthropic_system_prompt, anthropic_messages = get_anthropic_system_prompt_and_messages(messages)
            
            provider_data = get_anthropic_completion_function_params(
                model=self.model if response_format_info else ANTHROPIC_FAST_MODEL,
                messages=anthropic_messages,
                max_tokens=MAX_TOKENS,
                temperature=0,
                system=anthropic_system_prompt,
                stream=None,
                response_format_info=response_format_info
            )
            
            if self.api_key:
                # Unpack provider_data for direct API call
                assert self.client is not None
                response = self.client.messages.create(**provider_data)
                if provider_data.get("tool_choice") is not None:
                    result = extract_and_parse_anthropic_json_response(response)
                    return json.dumps(result) if not isinstance(result, str) else result
                else:
                    content = response.content
                    if content[0].type == "text":
                        return content[0].text
                    else:
                        return ""
            else:
                # Only pass provider_data to the server
                response = await get_anthropic_completion_from_mito_server(
                    model=provider_data["model"],
                    max_tokens=provider_data["max_tokens"],
                    temperature=provider_data["temperature"],
                    system=provider_data["system"],
                    messages=provider_data["messages"],
                    tools=provider_data.get("tools"),
                    tool_choice=provider_data.get("tool_choice"),
                    message_type=message_type
                )
                return response
        except anthropic.RateLimitError:
            raise Exception("Rate limit exceeded. Please try again later or reduce your request frequency.")
        except Exception as e:
            return f"Error streaming content: {str(e)}"

    async def stream_response(self, messages: List[ChatCompletionMessageParam], message_id: str, message_type: MessageType,
                              reply_fn: Callable[[Union[CompletionReply, CompletionStreamChunk]], None]) -> str:
        try:
            anthropic_system_prompt, anthropic_messages = get_anthropic_system_prompt_and_messages(messages)
            accumulated_response = ""

            if self.api_key:
                assert self.client is not None
                stream = self.client.messages.create(
                    model=self.model,
                    max_tokens=MAX_TOKENS,
                    temperature=0,
                    system=anthropic_system_prompt,
                    messages=anthropic_messages,
                    stream=True
                )


                for chunk in stream:
                    if chunk.type == "content_block_delta" and chunk.delta.type == "text_delta":
                        content = chunk.delta.text
                        accumulated_response += content

                        is_finished = chunk.type == "message_stop"

                        reply_fn(CompletionStreamChunk(
                            parent_id=message_id,
                            chunk=CompletionItem(
                                content=content,
                                isIncomplete=not is_finished,
                                token=message_id,
                            ),
                            done=is_finished,
                        ))

            else:
                async for stram_chunk in stream_anthropic_completion_from_mito_server(
                    model=self.model,
                    max_tokens=MAX_TOKENS,
                    temperature=0,
                    system=anthropic_system_prompt,
                    messages=anthropic_messages,
                    stream=True,
                    message_type=message_type
                ):
                    accumulated_response += stram_chunk
                    reply_fn(CompletionStreamChunk(
                        parent_id=message_id,
                        chunk=CompletionItem(
                            content=stram_chunk,
                            isIncomplete=True,
                            token=message_id,
                        ),
                        done=False,
                    ))

            return accumulated_response

        except anthropic.RateLimitError:
            raise Exception("Rate limit exceeded. Please try again later or reduce your request frequency.")

        except Exception as e:
            return f"Error streaming content: {str(e)}"





# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import json
import os
from typing import Any, Final, Callable
from functools import wraps
from mito_ai.utils.schema import MITO_FOLDER

SETTINGS_PATH: Final[str] = os.path.join(MITO_FOLDER, "settings.json")


def set_settings_field(field: str, value: Any) -> None:
    """
    Updates the value of a specific field in settings.json
    """
    with open(SETTINGS_PATH, "r") as user_file_old:
        old_user_json = json.load(user_file_old)
        old_user_json[field] = value
        with open(SETTINGS_PATH, "w+") as f:
            f.write(json.dumps(old_user_json))


def get_settings_field(field: str) -> Any:
    """
    Retrieves the value of a specific field from settings.json
    """
    with open(SETTINGS_PATH, "r") as user_file_old:
        return json.load(user_file_old).get(field)


def ensure_settings_file_exists(method: Callable) -> Callable:
    """Decorator to ensure the settings.json file exists before executing the handler method."""

    @wraps(method)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        if not os.path.exists(SETTINGS_PATH):
            with open(SETTINGS_PATH, "w") as f:
                json.dump({}, f, indent=4)
        return method(self, *args, **kwargs)

    return wrapper



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import List
from mito_ai.completions.models import AgentSmartDebugMetadata
from mito_ai.completions.prompt_builders.prompt_constants import (
    FILES_SECTION_HEADING,
    JUPYTER_NOTEBOOK_SECTION_HEADING,
    VARIABLES_SECTION_HEADING
)

# TODO:
# 1. In the future, it might make sense to pass the previous CELL_UPDATE to this prompt?
# 2. In the future, we should let the agent fix up the error by updating a different cell. This is sometimes a better solution. 
# However, to do this, we then need to know which code cells to run in order to validate the update is correct! If the error was 
# produced by code cell 3, and the agent corrects the source of the error in code cell 2, we then need to run cell 2 and 3 to validate
# the cell update worked properly. This could be many cells if there are intermediate cells. It might require something like a dependency 
# graph of cells that we calculate ourselves, not relying on the AI. 

def create_agent_smart_debug_prompt(md: AgentSmartDebugMetadata) -> str:
    variables_str = '\n'.join([f"{variable}" for variable in md.variables or []])
    files_str = '\n'.join([f"{file}" for file in md.files or []])
    ai_optimized_cells_str = '\n'.join([f"{cell}" for cell in md.aiOptimizedCells or []])
    
    return f"""I just applied and executed the CELL_UPDATE that you just shared with me, but it errored. Below I am sharing with you a strategy for how I want you to resolve this error and information about the actual error that occured.

Use this strategy for this message only. After this message, continue using the original set of instructions that I provided you.

It is very important that When fixing this error, you do not change the original intent of the code cell. 

To fix this error, take the following approach: 
Step 1: ERROR ANALYSIS: Analyze the error message to identify why the code cell errored.
Step 2: INTENT PRESERVATION: Make sure you understand the intent of the CELL_UPDATE so that you can be sure to preserve it when you create a new CELL_UPDATE
Step 3: ERROR CORRECTION: Respond with a new CELL_UPDATE that is applied to the same cell as the erroring CELL_UPDATE.

<Instructions for each Phase />

ERROR ANALYSIS:

- Identify error type (Syntax, Runtime, Logic).
- Use the defined variables and Jupyter Notebook to understand the error.
- Consider kernel state and execution order

INTENT PRESERVATION:

- Try to understand the user's intent using the defined variables and the Jupyter Notebook

ERROR CORRECTION:

- Return the full, updated version of cell {md.error_message_producing_code_cell_id} with the error fixed and a short explanation of the error.
- You can only update code in {md.error_message_producing_code_cell_id}. You are unable to edit the code in any other cell when resolving this error.
- Propose a solution that fixes the error and does not change the user's intent.
- Make the solution as simple as possible.
- Reuse as much of the existing code as possible.
- DO NOT ADD TEMPORARY COMMENTS like '# Fixed the typo here' or '# Added this line to fix the error'
- If you encounter a ModuleNotFoundError, you can install the package by adding the the following line to the top of the code cell: `!pip install <package_name> --quiet`.

<Example>

<Input>

{FILES_SECTION_HEADING}
file_name: sales.csv

Jupyter Notebook:
[
    {{
        cell_type: 'markdown'
        id: '9e38c62b-38f8-457d-bb8d-28bfc52edf2c'
        code: \"\"\"# Transaction Analysis \"\"\"
    }},
    {{
        cell_type: 'code'
        id: 'adslkaf-jf73-l8xn-92j7-kjd8kdcnd2kso'
        code: \"\"\" 'df' = pd.DataFrame({{
    'order_id': [1, 2, 3, 4],
    'date': ['Mar 7, 2025', 'Sep 24, 2024', '25 June, 2024', 'June 29, 2024'],
    'amount': [100, 150, 299, 99]
}})
    }},
    {{
        cell_type: 'code'
        id: 'c68fdf19-db8c-46dd-926f-d90ad35bb3bc'
        code: \"\"\"df['date'] = pd.to_datetime(df['date'])\"\"\"
    }},
]

{VARIABLES_SECTION_HEADING}
{{
    'df': pd.DataFrame({{
        'order_id': [1, 2, 3, 4],
        'date': ['Mar 7, 2025', 'Sep 24, 2024', '25 June, 2024', 'June 29, 2024'],
        'amount': [100, 150, 299, 99]
    }})
}}

Cell ID of the Error Producing Code Cell:
'c68fdf19-db8c-46dd-926f-d90ad35bb3bc'

Error Traceback:
Cell In[27], line 1
----> 1 df['date'] = pd.to_datetime(df['date'])

ValueError: time data "25 June, 2024" doesn't match format "%b %d, %Y", at position 2. You might want to try:
    - passing `format` if your strings have a consistent format;
    - passing `format='ISO8601'` if your strings are all ISO8601 but not necessarily in exactly the same format;
    - passing `format='mixed'`, and the format will be inferred for each element individually. You might want to use `dayfirst` alongside this.


</ Input>

< Your Thinking >

ERROR ANALYSIS
This is a ValueError caused by applying the wrong format to a specific date string. Because it was triggered at position 2, the first date string must have successfully converted. By looking at the defined variables, I can see that first date string is in the format "Mar 7, 2025", but the third date string is in the format "25 June, 2024". Those dates are not in the same format, so the conversion failed.

INTENT PRESERVATION:
User is trying to convert the date column to a datetime object even though the dates are not in the same starting format. 

</ Your Thinking >

<Output>


{{
    is_finished: false, 
    cell_update: {{
        type: 'modification'
        id: 'c68fdf19-db8c-46dd-926f-d90ad35bb3bc'
        code: "def parse_date(date_str):\n    formats = ['%b %d, %Y', '%d %B, %Y']\n\n    for fmt in formats:\n        try:\n            return pd.to_datetime(date_str, format=fmt)\n        except ValueError:\n            # Try next format\n            continue\n\n    # If not format worked, return Not a Time\n    return pd.NaT\n\ndf['date'] = df['date'].apply(lambda x: parse_date(x))"
    }}
}}

</Output>

</Example>

{FILES_SECTION_HEADING}
{files_str}

{JUPYTER_NOTEBOOK_SECTION_HEADING}
{ai_optimized_cells_str}

{VARIABLES_SECTION_HEADING}
{variables_str}

Cell ID of the Error Producing Code Cell:
{md.error_message_producing_code_cell_id}

Error Traceback:
{md.errorMessage}
"""



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from __future__ import annotations
from typing import Any, Callable, Dict, List, Optional, Union, cast
from mito_ai import constants
from openai.types.chat import ChatCompletionMessageParam
from traitlets import Instance, Unicode, default, validate
from traitlets.config import LoggingConfigurable
from openai.types.chat import ChatCompletionMessageParam

from mito_ai import constants
from mito_ai.enterprise.utils import is_azure_openai_configured
from mito_ai.gemini_client import GeminiClient
from mito_ai.openai_client import OpenAIClient
from mito_ai.anthropic_client import AnthropicClient
from mito_ai.logger import get_logger
from mito_ai.completions.models import (
    AICapabilities,
    CompletionError,
    CompletionItem,
    CompletionItemError,
    CompletionReply,
    CompletionStreamChunk,
    MessageType,
    ResponseFormatInfo, CompletionItemError,
)
from mito_ai.utils.telemetry_utils import (
    KEY_TYPE_PARAM,
    MITO_AI_COMPLETION_ERROR,
    MITO_SERVER_KEY,
    USER_KEY,
    log,
    log_ai_completion_success,
)
from mito_ai.constants import get_model_provider

__all__ = ["OpenAIProvider"]

class OpenAIProvider(LoggingConfigurable):
    """Provide AI feature through OpenAI services."""

    api_key = Unicode(
        config=True,
        allow_none=True,
        help="OpenAI API key. Default value is read from the OPENAI_API_KEY environment variable.",
    )

    last_error = Instance(
        CompletionError,
        allow_none=True,
        help="""Last error encountered when using the OpenAI provider.

This attribute is observed by the websocket provider to push the error to the client.""",
    )

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        config = kwargs.get('config', {})
        if 'api_key' in kwargs:
            config['OpenAIClient'] = {'api_key': kwargs['api_key']}
        kwargs['config'] = config

        super().__init__(log=get_logger(), **kwargs)
        self.last_error = None
        self._openai_client: Optional[OpenAIClient] = OpenAIClient(**config)

    @property
    def capabilities(self) -> AICapabilities:
        if constants.CLAUDE_API_KEY and not self.api_key:
            return AICapabilities(
                configuration={"model": "<dynamic>"},
                provider="Claude",
            )
        if constants.GEMINI_API_KEY and not self.api_key:
            return AICapabilities(
                configuration={"model": "<dynamic>"},
                provider="Gemini",
            )
        if self._openai_client:
            return self._openai_client.capabilities
        return AICapabilities(
            configuration={"model": "<dynamic>"},
            provider="Mito server",
        )

    @property
    def key_type(self) -> str:
        if constants.CLAUDE_API_KEY and not self.api_key:
            return "claude"
        if constants.GEMINI_API_KEY and not self.api_key:
            return "gemini"
        if self._openai_client:
            return self._openai_client.key_type
        return MITO_SERVER_KEY

    async def request_completions(
        self,
        message_type: MessageType,
        messages: List[ChatCompletionMessageParam],
        model: str,
        response_format_info: Optional[ResponseFormatInfo] = None,
        user_input: Optional[str] = None,
        thread_id: Optional[str] = None
    ) -> str:
        """
        Request completions from the AI provider.
        """
        self.last_error = None
        completion = None
        last_message_content = str(messages[-1].get('content', '')) if messages else ""
        model_type = get_model_provider(model)
        try:
            if model_type == "claude":
                api_key = constants.CLAUDE_API_KEY
                anthropic_client = AnthropicClient(api_key=api_key, model=model)
                completion = await anthropic_client.request_completions(messages, response_format_info, message_type)
            elif model_type == "gemini":
                api_key = constants.GEMINI_API_KEY
                gemini_client = GeminiClient(api_key=api_key, model=model)
                messages_for_gemini = [dict(m) for m in messages]
                completion = await gemini_client.request_completions(messages_for_gemini, response_format_info, message_type)
            elif model_type == "openai":
                if not self._openai_client:
                    raise RuntimeError("OpenAI client is not initialized.")
                completion = await self._openai_client.request_completions(
                    message_type=message_type,
                    messages=messages,
                    model=model,
                    response_format_info=response_format_info
                )
            else:
                raise ValueError(f"No AI provider configured for model: {model}")
            log_ai_completion_success(
                key_type=USER_KEY if self.key_type == "user" else MITO_SERVER_KEY,
                message_type=message_type,
                last_message_content=last_message_content,
                response={"completion": completion},
                user_input=user_input or "",
                thread_id=thread_id or "",
                model=model
            )
            return completion

        except BaseException as e:
            self.log.exception(f"Error during request_completions: {e}")
            self.last_error = CompletionError.from_exception(e)
            log(MITO_AI_COMPLETION_ERROR, params={KEY_TYPE_PARAM: self.key_type}, error=e)
            raise

    async def stream_completions(
        self,
        message_type: MessageType,
        messages: List[ChatCompletionMessageParam],
        model: str,
        message_id: str,
        thread_id: str,
        reply_fn: Callable[[Union[CompletionReply, CompletionStreamChunk]], None],
        user_input: Optional[str] = None,
        response_format_info: Optional[ResponseFormatInfo] = None
    ) -> str:
        """
        Stream completions from the AI provider and return the accumulated response.
        Returns: The accumulated response string.
        """
        self.last_error = None
        accumulated_response = ""
        last_message_content = str(messages[-1].get('content', '')) if messages else ""
        model_type = get_model_provider(model)
        reply_fn(CompletionReply(
            items=[
                CompletionItem(content="", isIncomplete=True, token=message_id)
            ],
            parent_id=message_id,
        ))

        try:
            if model_type == "claude":
                api_key = constants.CLAUDE_API_KEY
                anthropic_client = AnthropicClient(api_key=api_key, model=model)
                accumulated_response = await anthropic_client.stream_response(
                    messages=messages,
                    message_type=message_type,
                    message_id=message_id,
                    reply_fn=reply_fn
                )
            elif model_type == "gemini":
                api_key = constants.GEMINI_API_KEY
                gemini_client = GeminiClient(api_key=api_key, model=model)
                messages_for_gemini = [dict(m) for m in messages]
                accumulated_response = await gemini_client.stream_completions(
                    messages=messages_for_gemini,
                    message_id=message_id,
                    reply_fn=reply_fn,
                    message_type=message_type
                )
            elif model_type == "openai":
                if not self._openai_client:
                    raise RuntimeError("OpenAI client is not initialized.")
                accumulated_response = await self._openai_client.stream_completions(
                    message_type=message_type,
                    messages=messages,
                    model=model,
                    message_id=message_id,
                    thread_id=thread_id,
                    reply_fn=reply_fn,
                    user_input=user_input,
                    response_format_info=response_format_info
                )
            else:
                raise ValueError(f"No AI provider configured for model: {model}")

            # Log the successful completion
            log_ai_completion_success(
                key_type=USER_KEY if self.key_type == "user" else MITO_SERVER_KEY,
                message_type=message_type,
                last_message_content=last_message_content,
                response={"completion": accumulated_response},
                user_input=user_input or "",
                thread_id=thread_id,
                model=model
            )
            return accumulated_response

        except BaseException as e:
            self.log.exception(f"Error during stream_completions: {e}")
            self.last_error = CompletionError.from_exception(e)
            log(
                MITO_AI_COMPLETION_ERROR,
                params={
                    KEY_TYPE_PARAM: self.key_type,
                    'message_type': message_type.value,
                },
                error=e
            )
            # Send error message to client before raising
            reply_fn(CompletionStreamChunk(
                parent_id=message_id,
                chunk=CompletionItem(
                    content="",
                    isIncomplete=True,
                    error=CompletionItemError(
                        message=f"Failed to process completion: {e!r}"
                    ),
                    token=message_id,
                ),
                done=True,
                error=CompletionError.from_exception(e),
            ))
            raise




# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import pytest
from typing import Callable, List, cast
from openai.types.chat import ChatCompletionMessageParam
from mito_ai.utils.message_history_utils import trim_sections_from_message_content, trim_old_messages
from mito_ai.completions.prompt_builders.chat_prompt import create_chat_prompt
from mito_ai.completions.prompt_builders.agent_execution_prompt import create_agent_execution_prompt
from mito_ai.completions.prompt_builders.agent_smart_debug_prompt import (
    create_agent_smart_debug_prompt,
)
from mito_ai.completions.prompt_builders.smart_debug_prompt import create_error_prompt
from mito_ai.completions.prompt_builders.explain_code_prompt import create_explain_code_prompt
from mito_ai.completions.models import (
    AgentExecutionMetadata,
    AgentSmartDebugMetadata,
    AIOptimizedCell,
    ThreadID,
)
from mito_ai.completions.prompt_builders.prompt_constants import (
    FILES_SECTION_HEADING,
    VARIABLES_SECTION_HEADING,
    CODE_SECTION_HEADING,
    ACTIVE_CELL_ID_SECTION_HEADING,
    JUPYTER_NOTEBOOK_SECTION_HEADING,
    CONTENT_REMOVED_PLACEHOLDER,
)

# Standard test data for multiple tests
TEST_VARIABLES = ["'df': pd.DataFrame({'col1': [1, 2, 3], 'col2': [4, 5, 6]})"]
TEST_FILES = ["data.csv", "script.py"]
TEST_CODE = "import pandas as pd\ndf = pd.read_csv('data.csv')"
TEST_INPUT = "Calculate the mean of col1"
TEST_ERROR = "AttributeError: 'Series' object has no attribute 'mena'"

def test_trim_sections_basic() -> None:
    """Test trimming sections on a simple string with all section types."""
    content = f"""Some text before.

{FILES_SECTION_HEADING}
file1.csv
file2.txt
file3.py

{VARIABLES_SECTION_HEADING}
var1 = 1
var2 = "string"
var3 = [1, 2, 3]

{JUPYTER_NOTEBOOK_SECTION_HEADING}
[
    {{
        "cell_type": "code",
        "id": "cell1",
        "code": "print('hello world')"
    }}
]

{ACTIVE_CELL_ID_SECTION_HEADING}
cell1

{CODE_SECTION_HEADING}
```python
def hello():
    print("world")
```

Some text after."""

    result = trim_sections_from_message_content(content)

    # Verify sections are replaced with placeholders
    assert f"{FILES_SECTION_HEADING} {CONTENT_REMOVED_PLACEHOLDER}" in result
    assert f"{VARIABLES_SECTION_HEADING} {CONTENT_REMOVED_PLACEHOLDER}" in result
    assert f"{JUPYTER_NOTEBOOK_SECTION_HEADING} {CONTENT_REMOVED_PLACEHOLDER}" in result
    assert f"{ACTIVE_CELL_ID_SECTION_HEADING} {CONTENT_REMOVED_PLACEHOLDER}" in result
    
    # Verify sections are not in the result anymore
    assert "file1.csv" not in result
    assert "var1 = 1" not in result
    assert "cell_type" not in result
    assert "cell1" not in result
    
    # Verify other content is preserved
    assert "Some text before." in result
    assert "Some text after." in result
    assert f"{CODE_SECTION_HEADING}" in result
    assert "def hello():" in result


# Parameterized test cases for prompt builders (except inline completer)
PROMPT_BUILDER_TEST_CASES = [
    # Chat prompt
    (
        lambda: create_chat_prompt(TEST_VARIABLES, TEST_FILES, TEST_CODE, "cell1", False, TEST_INPUT),
        ["Your task: Calculate the mean of col1"],
        ["data.csv\nscript.py", f"{VARIABLES_SECTION_HEADING}\n'df': pd.DataFrame"],
    ),
    # Agent execution prompt
    (
        lambda: create_agent_execution_prompt(
            AgentExecutionMetadata(
                variables=TEST_VARIABLES,
                files=TEST_FILES,
                aiOptimizedCells=[
                    AIOptimizedCell(cell_type="code", id="cell1", code=TEST_CODE)
                ],
                input=TEST_INPUT,
                promptType="agent:execution",
                threadId=ThreadID("test-thread-id"),
                isChromeBrowser=True
            )
        ),
        ["Your task: \nCalculate the mean of col1"],
        ["data.csv\nscript.py", f"'df': pd.DataFrame", "import pandas as pd"],
    ),
    # Smart debug prompt
    (
        lambda: create_error_prompt(TEST_ERROR, TEST_CODE, "cell1", TEST_VARIABLES, TEST_FILES),
        ["Error Traceback:", TEST_ERROR],
        [
            f"{FILES_SECTION_HEADING}\ndata.csv\nscript.py",
            f"{VARIABLES_SECTION_HEADING}\n'df': pd.DataFrame",
            f"{ACTIVE_CELL_ID_SECTION_HEADING}\ncell1",
        ],
    ),
    # Explain code prompt (doesn't have sections to trim)
    (
        lambda: create_explain_code_prompt(TEST_CODE),
        ["import pandas as pd"],
        [],
    ),
    # Agent smart debug prompt
    (
        lambda: create_agent_smart_debug_prompt(
            AgentSmartDebugMetadata(
                variables=TEST_VARIABLES,
                files=TEST_FILES,
                aiOptimizedCells=[
                    AIOptimizedCell(cell_type="code", id="cell1", code=TEST_CODE)
                ],
                error_message_producing_code_cell_id="cell1",
                errorMessage=TEST_ERROR,
                promptType="agent:autoErrorFixup",
                threadId=ThreadID("test-thread-id"),
                isChromeBrowser=True
            )
        ),
        ["Error Traceback:", TEST_ERROR],
        [
            f"{FILES_SECTION_HEADING}\n{TEST_FILES[0]}",
            f"{VARIABLES_SECTION_HEADING}\n{TEST_VARIABLES[0]}",
            "cell_type",
        ],
    ),
]


@pytest.mark.parametrize("prompt_builder,expected_in_result,expected_not_in_result", PROMPT_BUILDER_TEST_CASES)
def test_prompt_builder_trimming(prompt_builder: Callable[[], str], expected_in_result: List[str], expected_not_in_result: List[str]) -> None:
    """Test trimming for different prompt builders."""
    # Create prompt using the provided builder function
    content = prompt_builder()

    # Trim the content
    result = trim_sections_from_message_content(content)

    # If none of the section headings are present, the content shouldn't change
    has_sections_to_trim = any(
        heading in content 
        for heading in [FILES_SECTION_HEADING, VARIABLES_SECTION_HEADING, JUPYTER_NOTEBOOK_SECTION_HEADING, ACTIVE_CELL_ID_SECTION_HEADING]
    )
    
    if not has_sections_to_trim:
        assert result == content
        # Verify expected content is still in the result
        for expected in expected_in_result:
            assert expected in result
        return

    # Check for each section if it was present in the original content
    if FILES_SECTION_HEADING in content:
        assert f"{FILES_SECTION_HEADING} {CONTENT_REMOVED_PLACEHOLDER}" in result
        
    if VARIABLES_SECTION_HEADING in content:
        assert f"{VARIABLES_SECTION_HEADING} {CONTENT_REMOVED_PLACEHOLDER}" in result
        
    if JUPYTER_NOTEBOOK_SECTION_HEADING in content:
        assert (
            f"{JUPYTER_NOTEBOOK_SECTION_HEADING} {CONTENT_REMOVED_PLACEHOLDER}"
            in result
        )

    if ACTIVE_CELL_ID_SECTION_HEADING in content:
        assert f"{ACTIVE_CELL_ID_SECTION_HEADING} {CONTENT_REMOVED_PLACEHOLDER}" in result

    # Verify expected content is still in the result
    for expected in expected_in_result:
        assert expected in result

    # Verify content that should be removed is not in the result
    for not_expected in expected_not_in_result:
        assert not_expected not in result

def test_no_sections_to_trim() -> None:
    """Test trimming content with no sections to trim."""
    content = "This is a simple message with no sections to trim."
    result = trim_sections_from_message_content(content)
    assert result == content


# Tests for trim_old_messages function
def test_trim_old_messages_only_trims_user_messages() -> None:
    """Test that trim_old_messages only trims content from user messages."""
    # Create test messages with different roles
    user_message_with_sections = f"""User prompt with sections.
{FILES_SECTION_HEADING}
file1.csv
file2.txt
"""
    system_message_with_sections = f"""System message with sections.
{FILES_SECTION_HEADING}
file1.csv
file2.txt
"""
    assistant_message_with_sections = f"""Assistant message with sections.
{FILES_SECTION_HEADING}
file1.csv
file2.txt
"""
    
    # Create test messages with proper typing
    messages: List[ChatCompletionMessageParam] = [
        {"role": "system", "content": system_message_with_sections},
        {"role": "user", "content": user_message_with_sections},
        {"role": "assistant", "content": assistant_message_with_sections},
        {"role": "user", "content": "Recent user message"},
    ]
    
    # Keep only the most recent message
    result = trim_old_messages(messages, keep_recent=1)
    
    # System message should remain unchanged even though it's old
    system_content = result[0].get("content")
    assert isinstance(system_content, str)
    assert system_content == system_message_with_sections
    assert FILES_SECTION_HEADING in system_content
    assert "file1.csv" in system_content
    
    # First user message should be trimmed
    assert result[1]["role"] == "user"
    user_content = result[1].get("content")
    assert isinstance(user_content, str)
    assert f"{FILES_SECTION_HEADING} {CONTENT_REMOVED_PLACEHOLDER}" in user_content
    assert "file1.csv" not in user_content
    
    # Assistant message should remain unchanged even though it's old
    assistant_content = result[2].get("content")
    assert isinstance(assistant_content, str)
    assert assistant_content == assistant_message_with_sections
    assert FILES_SECTION_HEADING in assistant_content
    assert "file1.csv" in assistant_content
    
    # Recent user message should remain unchanged
    recent_content = result[3].get("content")
    assert isinstance(recent_content, str)
    assert recent_content == "Recent user message"


def test_trim_old_messages_preserves_recent_messages() -> None:
    """Test that trim_old_messages preserves the most recent messages based on keep_recent parameter."""
    # Create test messages
    old_message_1 = f"""Old message 1.
{FILES_SECTION_HEADING}
file1.csv
"""
    old_message_2 = f"""Old message 2.
{FILES_SECTION_HEADING}
file2.csv
"""
    recent_message_1 = f"""Recent message 1.
{FILES_SECTION_HEADING}
file3.csv
"""
    recent_message_2 = f"""Recent message 2.
{FILES_SECTION_HEADING}
file4.csv
"""
    
    # Create test messages with proper typing
    messages: List[ChatCompletionMessageParam] = [
        {"role": "user", "content": old_message_1},
        {"role": "user", "content": old_message_2},
        {"role": "user", "content": recent_message_1},
        {"role": "user", "content": recent_message_2},
    ]
    
    # Keep the 2 most recent messages
    result = trim_old_messages(messages, keep_recent=2)
    
    # Old messages should be trimmed
    old_content_1 = result[0].get("content")
    assert isinstance(old_content_1, str)
    assert f"{FILES_SECTION_HEADING} {CONTENT_REMOVED_PLACEHOLDER}" in old_content_1
    assert "file1.csv" not in old_content_1
    
    old_content_2 = result[1].get("content")
    assert isinstance(old_content_2, str)
    assert f"{FILES_SECTION_HEADING} {CONTENT_REMOVED_PLACEHOLDER}" in old_content_2
    assert "file2.csv" not in old_content_2
    
    # Recent messages should remain unchanged
    recent_content_1 = result[2].get("content")
    assert isinstance(recent_content_1, str)
    assert recent_content_1 == recent_message_1
    assert FILES_SECTION_HEADING in recent_content_1
    assert "file3.csv" in recent_content_1
    
    recent_content_2 = result[3].get("content")
    assert isinstance(recent_content_2, str)
    assert recent_content_2 == recent_message_2
    assert FILES_SECTION_HEADING in recent_content_2
    assert "file4.csv" in recent_content_2

def test_trim_old_messages_empty_list() -> None:
    """Test that trim_old_messages handles empty message lists correctly."""
    messages: List[ChatCompletionMessageParam] = []
    result = trim_old_messages(messages, keep_recent=2)
    assert result == []


def test_trim_old_messages_fewer_than_keep_recent() -> None:
    """Test that trim_old_messages doesn't modify messages if there are fewer than keep_recent."""
    messages: List[ChatCompletionMessageParam] = [
        {"role": "user", "content": "User message 1"},
        {"role": "assistant", "content": "Assistant message 1"},
    ]
    
    result = trim_old_messages(messages, keep_recent=3)
    
    # Messages should remain unchanged
    user_content = result[0].get("content")
    assert isinstance(user_content, str)
    assert user_content == "User message 1"
    
    assistant_content = result[1].get("content")
    assert isinstance(assistant_content, str)
    assert assistant_content == "Assistant message 1"


def test_trim_mixed_content_messages() -> None:
    """
    Tests that when a message contains sections other than text (like image_url),
    those sections are removed completely, leaving only the text content.
    """
    # Create sample message with mixed content (text and image)
    mixed_content_message = cast(ChatCompletionMessageParam, {
        "role": "user",
        "content": [
            {
                "type": "text",
                "text": "What is in this image?"
            },
            {
                "type": "image_url",
                "image_url": {"url": "data:image/png;base64,someimagedata"}
            }
        ]
    })
    
    # Create sample message list with one old message (the mixed content)
    # and one recent message (to not be trimmed)
    message_list: List[ChatCompletionMessageParam] = [
        mixed_content_message,  # This should get trimmed
        {"role": "assistant", "content": "That's a chart showing data trends"},
        {"role": "user", "content": "Can you explain more?"}  # Recent message, should not be trimmed
    ]
    
    # Apply the trimming function
    trimmed_messages = trim_old_messages(message_list, keep_recent=2)
    
    # Verify that the first message has been trimmed properly
    assert trimmed_messages[0]["role"] == "user"
    assert trimmed_messages[0]["content"] == "What is in this image?"
    
    # Verify that the recent messages are untouched
    assert trimmed_messages[1] == message_list[1]
    assert trimmed_messages[2] == message_list[2]


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import Dict, TypedDict, List


class DatabaseConfig(TypedDict, total=False):
    drivers: List[str]
    tables_query: str
    columns_query: str


SUPPORTED_DATABASES: Dict[str, DatabaseConfig] = {
    "mssql": {
        "drivers": ["pyodbc"],
        "tables_query": "SELECT table_name FROM information_schema.tables WHERE table_schema = 'dbo'",
        "columns_query": "SELECT column_name, data_type FROM information_schema.columns WHERE table_name = :table",
    },
    "mysql": {
        "drivers": ["PyMySQL"],
        "tables_query": "SHOW TABLES",
        "columns_query": "SHOW COLUMNS FROM {table}",
    },
    "oracle": {
        "drivers": ["oracledb"],
        "tables_query": "SELECT table_name FROM user_tables",
        "columns_query": "SELECT column_name, data_type FROM user_tab_columns WHERE table_name = :table",
    },
    "postgres": {
        "drivers": ["psycopg2-binary"],
        "tables_query": "SELECT table_name FROM information_schema.tables WHERE table_schema = :schema",
        "columns_query": "SELECT column_name, data_type FROM information_schema.columns WHERE table_name = :table",
    },
    "snowflake": {
        "drivers": ["snowflake-sqlalchemy"],
        # Queries handled in the snowflake.py file.
    },
    "sqlite": {
        "drivers": [],
        "tables_query": "SELECT name FROM sqlite_master WHERE type='table'",
        "columns_query": "SELECT name, type FROM pragma_table_info(:table)",
    },
}



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

"""
This module contains constants used in prompts across the codebase.
These constants ensure consistency between prompt building and message trimming.
"""

import os
import json
from typing import Final
from mito_ai.utils.schema import MITO_FOLDER

# Section headings used in prompts
FILES_SECTION_HEADING = "Files in the current directory:"
VARIABLES_SECTION_HEADING = "Defined Variables:"
CODE_SECTION_HEADING = "Code in the active code cell:"
ACTIVE_CELL_ID_SECTION_HEADING = "The ID of the active code cell:"
ACTIVE_CELL_OUTPUT_SECTION_HEADING = "Output of the active code cell:"
GET_CELL_OUTPUT_TOOL_RESPONSE_SECTION_HEADING = "Output of the code cell you just applied the CELL_UPDATE to:"
JUPYTER_NOTEBOOK_SECTION_HEADING = "Jupyter Notebook:"

# Placeholder text used when trimming content from messages
CONTENT_REMOVED_PLACEHOLDER = "Content removed to save space" 

CITATION_RULES = """RULES FOR CITING YOUR WORK

It is important that the user is able to verify any insights that you share with them about their data. To make this easy for the user, you must cite the lines of code that you are drawing the insight from. To provide a citation, use one of the following formats inline in your response:

Single line citation:
[MITO_CITATION:cell_id:line_number]

Multiline citation (for citing a range of lines):
[MITO_CITATION:cell_id:first_line-last_line]

Citation Rules:

1. Every fact or statement derived from the user's notebook must include a citation. 
2. When choosing the citation, select the code that will most help the user validate the fact or statement that you shared with them.
3. Place the citation immediately after the statement it supports. Do not explain the citation with phrases like "See", "Derived from", etc. Just provide the citation object.
4. For the "line_number" field, use the line number within the cell that is most relevant to the citation. Important: The cell line number should be 0-indexed and should not skip comments.
5. For multiline citations, use the "first_line-last_line" format when the insight spans multiple lines of code. Both line numbers should be 0-indexed.
6. If you cannot find relevant information in the notebook to answer a question, clearly state this and do not provide a citation.
7. You ONLY need to provide a citation when sharing an insight from the data in the message part of the response. If all you are doing is writing/updating code, then there is no need to provide a citation.
8. Do not include the citation in the code block as a comment. ONLY include the citation in the message field of your response.
"""

def get_active_cell_output_str(has_active_cell_output: bool) -> str:
    """
    Used to tell the AI about the output of the active code cell. 
    We use this in the chat prompt.
    """
    if has_active_cell_output:
        return f"{ACTIVE_CELL_OUTPUT_SECTION_HEADING}\nAttatched is an image of the output of the active code cell for your context."
    else:
        return ""
    
def cell_update_output_str(has_cell_update_output: bool) -> str:
    """
    Used to respond to the GET_CELL_OUTPUT tool, telling the agent the output of the cell it requested
    """
    if has_cell_update_output:
        return f"{GET_CELL_OUTPUT_TOOL_RESPONSE_SECTION_HEADING}\nAttatched is an image of code cell output that you requested."
    else:
        return ""

def redact_sensitive_info(connections: dict) -> dict:
    """
    Redacts sensitive information from connections data.
    Returns a copy of the connections dict with sensitive fields masked.
    """
    redacted = {}
    for conn_name, conn_data in connections.items():
        redacted[conn_name] = conn_data.copy()
        for key, value in redacted[conn_name].items():
            redacted[conn_name][key] = 'redacted'
    return redacted

def get_database_rules() -> str:
    """
    Reads the user's database configurations,
    and returns the rules for the AI to follow.
    """

    # Get the db configuration from the user's mito folder

    APP_DIR_PATH: Final[str] = os.path.join(MITO_FOLDER)
    connections_path: Final[str] = os.path.join(APP_DIR_PATH, 'db', 'connections.json')
    schemas_path: Final[str] = os.path.join(APP_DIR_PATH, 'db', 'schemas.json')
    
    try:
        with open(connections_path, 'r') as f:
            connections = json.load(f)
            sanitized_connections = redact_sensitive_info(connections)
    except FileNotFoundError:
        connections = None
        sanitized_connections = None

    try:
        with open(schemas_path, 'r') as f:
            schemas = json.load(f)
    except FileNotFoundError:
        schemas = None

    # If there is a db configuration, add return the rules

    if connections is not None:
        DATABASE_RULES = f"""DATABASE RULES:
If the user has requested data that you believe is stored in the database:
- Use the provided schema.
- Only use SQLAlchemy to query the database.
- Do not use a with statement when creating the SQLAlchemy engine. Instead, initialize it once so it can be reused for multiple queries.
- Always return the results of the query in a pandas DataFrame, unless instructed otherwise.
- Every schema has a unique connection ID. This ID can be used to find the connection details in the connections.json file.
- Do not use the connection ID to query the database. It is only for matching the schema to the correct connection.
- When using the connection ID, do not include any comments about it in your code.
- Connection details are stored in a JSON file located at: `{connections_path}`
- Here is the sanitized contents of the connections.json file:

{sanitized_connections}

- Do not hard-code connection credentials into your code. Instead, load the connections.json file and access connection fields dynamically like so:

```
connections[connection_name]["username"]
```

- When connecting to an Oracle database, use the following format:
```
conn_str = f"oracle+oracledb://username:password@host:port?service_name=service_name"
```

- The user may colloquially ask for a "list of x", always assume they want a pandas DataFrame. 
- When working with dataframes created from an SQL query, ALWAYS use lowercase column names. 
- If you think the requested data is stored in the database, but you are unsure, then ask the user for clarification.

Here is the schema:
{schemas}
        """
    else:
        DATABASE_RULES = ""

    return DATABASE_RULES


CHAT_CODE_FORMATTING_RULES = """CRITICAL CODE UPDATE RULES:
- COMPLETE REPLACEMENT: Your code will COMPLETELY REPLACE the entire contents of the active code cell. 
- INCLUDE ALL CODE: You MUST return the COMPLETE, FULL contents of the entire code cell - including ALL existing code that should remain plus your modifications.
- NEVER PARTIAL CODE: NEVER return only a portion, snippet, or subset of the code cell. Partial responses will break the user's notebook by deleting important code.
- PRESERVE EXISTING CODE: Always preserve imports, variable definitions, and other code that the user needs, even if you're only modifying one small part.
"""


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import asyncio
import json
import time
from typing import Any, Dict, List, Optional, Callable, Union, AsyncGenerator, Tuple
from tornado.httpclient import AsyncHTTPClient
from mito_ai.completions.models import AgentResponse, CompletionReply, CompletionStreamChunk, CompletionItem, MessageType
from .utils import _create_http_client
from mito_ai.constants import MITO_GEMINI_URL

timeout = 30
max_retries = 1

def _prepare_gemini_request_data_and_headers(
    model: str,
    contents: List[Dict[str, Any]],
    message_type: MessageType,
    config: Optional[Dict[str, Any]] = None,
    response_format_info: Optional[Any] = None,
    stream: bool = False
) -> Tuple[Dict[str, Any], Dict[str, str]]:
    
    inner_data: Dict[str, Any] = {
        "model": model,
        "contents": contents,
        "message_type": message_type.value if hasattr(message_type, 'value') else str(message_type),
    }
    
    if response_format_info:
        # Ensure the format is a string, not a class
        format_value = getattr(response_format_info, 'format', None)
        if isinstance(format_value, type):
            format_value = format_value.__name__
        inner_data["response_format_info"] = json.dumps({
            "name": getattr(response_format_info, 'name', None),
            "format": format_value
        })
        
    if stream:
        inner_data["stream"] = True
        
    if config:
        # Ensure config is serializable
        inner_data["config"] = json.loads(json.dumps(config))
        
    data = {
        "timeout": timeout,
        "max_retries": max_retries,
        "data": inner_data
    }
    
    headers = {"Content-Type": "application/json"}
    return data, headers

async def get_gemini_completion_from_mito_server(
    model: str,
    contents: List[Dict[str, Any]],
    message_type: MessageType,
    config: Optional[Dict[str, Any]] = None,
    response_format_info: Optional[Any] = None
) -> str:
    data, headers = _prepare_gemini_request_data_and_headers(model, contents, message_type, config, response_format_info, stream=False)
    http_client, http_client_timeout = _create_http_client(timeout, max_retries)
    start_time = time.time()
    try:
        res = await http_client.fetch(
            MITO_GEMINI_URL,
            method="POST",
            headers=headers,
            body=json.dumps(data),
            request_timeout=http_client_timeout
        )
        print(f"Gemini request completed in {time.time() - start_time:.2f} seconds")
    except Exception as e:
        print(f"Gemini request failed after {time.time() - start_time:.2f} seconds with error: {str(e)}")
        raise
    finally:
        http_client.close()
    
    # The response is a string
    return res.body.decode("utf-8")

async def stream_gemini_completion_from_mito_server(
    model: str,
    contents: List[Dict[str, Any]],
    message_type: MessageType,
    message_id: str,
    reply_fn: Optional[Callable[[Union[CompletionReply, CompletionStreamChunk]], None]]
) -> AsyncGenerator[str, None]:
    data, headers = _prepare_gemini_request_data_and_headers(model, contents, message_type, stream=True)
    http_client, http_client_timeout = _create_http_client(timeout, max_retries)
    start_time = time.time()
    chunk_queue: asyncio.Queue[str] = asyncio.Queue()
    fetch_complete = False
    def chunk_callback(chunk: bytes) -> None:
        try:
            chunk_str = chunk.decode('utf-8')
            asyncio.create_task(chunk_queue.put(chunk_str))
        except Exception as e:
            print(f"Error processing Gemini streaming chunk: {str(e)}")
    fetch_future = None
    try:
        fetch_future = http_client.fetch(
            MITO_GEMINI_URL,
            method="POST",
            headers=headers,
            body=json.dumps(data),
            request_timeout=http_client_timeout,
            streaming_callback=chunk_callback
        )
        async def wait_for_fetch() -> None:
            try:
                await fetch_future
                nonlocal fetch_complete
                fetch_complete = True
                print("Gemini fetch completed")
            except Exception as e:
                print(f"Error in Gemini fetch: {str(e)}")
                raise
        fetch_task = asyncio.create_task(wait_for_fetch())
        while not (fetch_complete and chunk_queue.empty()):
            try:
                chunk = await asyncio.wait_for(chunk_queue.get(), timeout=0.1)
                clean_chunk = chunk.strip('"')
                decoded_chunk = clean_chunk.encode().decode('unicode_escape')
                if reply_fn and message_id:
                    reply_fn(CompletionStreamChunk(
                        parent_id=message_id,
                        chunk=CompletionItem(
                            content=decoded_chunk,
                            isIncomplete=True,
                            token=message_id,
                        ),
                        done=False,
                    ))
                yield chunk
            except asyncio.TimeoutError:
                if fetch_complete and chunk_queue.empty():
                    break
                continue
        print(f"\nGemini stream completed in {time.time() - start_time:.2f} seconds")
        if reply_fn and message_id:
            reply_fn(CompletionStreamChunk(
                parent_id=message_id,
                chunk=CompletionItem(
                    content="",
                    isIncomplete=False,
                    token=message_id,
                ),
                done=True,
            ))
    except Exception as e:
        print(f"\nGemini stream failed after {time.time() - start_time:.2f} seconds with error: {str(e)}")
        if fetch_future:
            try:
                await fetch_future
            except Exception:
                pass
        raise
    finally:
        http_client.close()

def get_gemini_completion_function_params(
    model: str,
    contents: list[dict[str, Any]],
    message_type: MessageType,
    response_format_info: Optional[Any] = None,
) -> Dict[str, Any]:
    """
    Build the provider_data dict for Gemini completions, mirroring the OpenAI/Anthropic approach.
    Only includes fields needed for the Gemini API.
    """
    provider_data: Dict[str, Any] = {
        "model": model,
        "contents": contents,
        "message_type": message_type.value if hasattr(message_type, 'value') else str(message_type),
    }
        
    # Configure response format if provided
    if response_format_info:
        provider_data["config"] = {
            "response_mime_type": "application/json",
            "response_schema": AgentResponse.model_json_schema()
        }
        
    return provider_data


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from mito_ai.completions.models import AgentExecutionMetadata
from mito_ai.completions.prompt_builders.prompt_constants import (
    FILES_SECTION_HEADING,
    JUPYTER_NOTEBOOK_SECTION_HEADING,
    VARIABLES_SECTION_HEADING,
    cell_update_output_str
)
from mito_ai.completions.prompt_builders.utils import get_rules_str

def create_agent_execution_prompt(md: AgentExecutionMetadata) -> str:
    variables_str = '\n'.join([f"{variable}" for variable in md.variables or []])
    files_str = '\n'.join([f"{file}" for file in md.files or []])
    ai_optimized_cells_str = '\n'.join([f"{cell}" for cell in md.aiOptimizedCells or []])
    rules_str = get_rules_str(md.selectedRules)
    
    context_str = f"""Remember to choose the correct tool to respond with.

{rules_str}


{JUPYTER_NOTEBOOK_SECTION_HEADING}
{ai_optimized_cells_str}

{VARIABLES_SECTION_HEADING}
{variables_str}

{FILES_SECTION_HEADING}
{files_str}

{cell_update_output_str(md.base64EncodedActiveCellOutput is not None)}"""

    task_str = '' if md.input == '' else f"""Your task: 
{md.input}"""

    return '\n\n'.join([context_str, task_str]).strip()


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import requests
from mito_ai.tests.db.test_db_constants import SNOWFLAKE
from mito_ai.tests.conftest import TOKEN


# --- ADD CONNECTION ---


def test_add_connection_with_auth(jp_base_url: str) -> None:
    response = requests.post(
        jp_base_url + "/mito-ai/db/connections",
        headers={"Authorization": f"token {TOKEN}"},
        json=SNOWFLAKE,
    )
    if response.status_code != 200:
        print(f"Error response: {response.text}")
    assert response.status_code == 200


def test_add_connection_with_no_auth(jp_base_url: str) -> None:
    response = requests.post(
        jp_base_url + "/mito-ai/db/connections",
        json=SNOWFLAKE,
    )
    assert response.status_code == 403  # Forbidden


def test_add_connection_with_incorrect_auth(jp_base_url: str) -> None:
    response = requests.post(
        jp_base_url + "/mito-ai/db/connections",
        headers={"Authorization": f"token incorrect-token"},
        json=SNOWFLAKE,
    )
    assert response.status_code == 403  # Forbidden


# --- GET CONNECTIONS ---


def test_get_connections_with_auth(jp_base_url: str, first_connection_id: str) -> None:
    response = requests.get(
        jp_base_url + "/mito-ai/db/connections",
        headers={"Authorization": f"token {TOKEN}"},
    )
    assert response.status_code == 200
    # Ensure the response has the correct number of connections
    assert len(response.json()) == 1
    # Ensure the response has the correct connection details
    assert response.json()[first_connection_id] == SNOWFLAKE


def test_get_connections_with_no_auth(jp_base_url: str) -> None:
    response = requests.get(jp_base_url + "/mito-ai/db/connections")
    assert response.status_code == 403  # Forbidden


def test_get_connections_with_incorrect_auth(jp_base_url: str) -> None:
    response = requests.get(
        jp_base_url + "/mito-ai/db/connections",
        headers={"Authorization": f"token incorrect-token"},
    )
    assert response.status_code == 403  # Forbidden


# --- DELETE CONNECTION ---


def test_delete_connection_with_no_auth(jp_base_url: str, first_connection_id: str) -> None:
    response = requests.delete(
        jp_base_url + f"/mito-ai/db/connections/{first_connection_id}",
    )
    assert response.status_code == 403  # Forbidden


def test_delete_connection_with_incorrect_auth(
    jp_base_url: str, first_connection_id: str
) -> None:
    response = requests.delete(
        jp_base_url + f"/mito-ai/db/connections/{first_connection_id}",
        headers={"Authorization": f"token incorrect-token"},
    )
    assert response.status_code == 403  # Forbidden


def test_delete_connection_with_auth(
    jp_base_url: str, first_connection_id: str
) -> None:
    response = requests.delete(
        jp_base_url + f"/mito-ai/db/connections/{first_connection_id}",
        headers={"Authorization": f"token {TOKEN}"},
    )
    assert response.status_code == 200

    # Ensure the connection was deleted
    response = requests.get(
        jp_base_url + f"/mito-ai/db/connections",
        headers={"Authorization": f"token {TOKEN}"},
    )
    assert len(response.json()) == 0
